# pigs.py
#
# PiGS Channel Master CM-7500 DVR+ Guide Server Replacement
#
#---------------------------------------------------------------------------------------
# Copyright 2021,2022, 2023 Timothy Goihl
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# See GNU-GPL-V3.txt License File
#
#---------------------------------------------------------------------------------------
# Files in the PiGS Guide Server project
#  pigs.py The main application / server
#  ls.py   PiGS LinearSchedule & Build functions
#  sv.py   PiGS Service Details functions
#  cm.py   Channel Master / Rovi APIs (access CM/Rovi servers) to get image data
#  sd.py   Schedules Direct.org API calls
#  ut.py   Utility functions that may be used by more than one module
#  se.py   Searching schedule from url
#  find.htm    Simple search UI for se.py
#  index.htm   Configuration UI for pigs. static IP e.g. //192.168.1.200
#  pigsdns.py  DNS server for DVR+ added in v1.02a
#
#  5 image files favicon.ico, pinkpig.png, 92x36box.png, misspiggy40x30.png, pigoutline92x36.png
#  pigs.service   a file that can be used to start pigs at boot (ie Auto-Run)
#  pigs.desktop   a file that can be placed on the Pi desktop to manually start PiGS
#  pigsreq.txt    requirements file for installing 3 dependancies
#    sudo pip3 install -r pigsreq.txt
#
#  GNU-GPL-V3.txt License File
#  ReleaseNotes.txt - verbose history of version changes
#  version.txt   a file placed on the pigs download server so Pigs can check for new updates
#   invoke nano with -L to prevent it from adding newline <LF> on save
#
#  scan.py Test functions to CM/Rovi servers, also a Simple file server.
#     scan.py needs cm.py to run, but that is it.
#     see instructions at the top of scan.py
#
#---------------------------------------------------------------------------------------
# To run pigs.py from a Terminal window, type:
#    sudo python3 pigs.py 
# Ctrl-C to exit/stop server
# Sudo is required since PiGS uses low level ports 80 & 53 for web & dns servers
#
# Companion Spreadsheet with install and development Notes.
# https://docs.google.com/spreadsheets/d/1fg2dEN390CaZ0fP0DSSJjdO8EyvbOL_oaIsbLogmE8I/edit?usp=sharing
# wordpress site for pigs updates: http://pigs.goihl.com   
# AVS Forum for PiGS: www.avsforum.com/threads/pigs-the-channel-master-dvr-ep-guide-server-replacement.3155354
#---------------------------------------------------------------------------------------

PIGS_VERSION = "V1.40 2024-Jan-14"   # must be this format for version update check

# early test to see if user installed the required modules

import sys

try: 
    from dateutil import parser
    import requests
    from PIL import Image

except:
    print("\nPiGS Error!  Could not find one or more of the required libraries!\n" )
    print("From a terminal window change to the PiGS folder then")
    print("install required libaries with the following command:")
    print("  sudo pip3 install -r pigsreq.txt   (if pip3 doesnt work, just 'pip' might)\n")
    print("If 'pip3' is not found, then install pip3 first:")
    print("  sudo apt-get install python3-pip")
    # Exit pigs, no sense in running pigs if required libraries not installed
    sys.exit(5)

    
# See ReleaseNotes.txt for detailed version history
#
# ToDo's
# ToDo's make a help.htm page that covers the whole UI
# Make a user.json file with setting that user can change
# make SD calls return  a dict of info instead of 3 params
# Aug 21, SD requires a token as a URL parameter to fetch show images (not chan logos)
# Test if Builder thread is alive on web request from PC or DVR, send email, restart?
# Expand 'Mark as new' concept...  keep a list of recorded episodes, set as new if not already recorded
#  Maybe call it RecordSeries
#  eg America's Test Kitchen repeats episodes multiple time a week/day like 4x or 5x
#    EP005502520424. "Elegant Dinner Party." S19, E7.
#  we can use the EPxxx number to mark it as new just 1x
# On SaveSetting do a json diff to see if only DNS has changed, no build if so
#   https://towardsdev.com/how-to-compare-and-find-difference-between-two-json-object-in-python-22cf86b7da3b
#   pip3 install deepdiff
#   from deepdiff import DeepDiff
#   result = DeepDiff(js1, js2)
#   'dictionary_item_added': [], dictionary_item_removed': [], dictionary_item_changed': [], 
# On startup Maybe scan saved dvr list to set 1st build to 1hr before
# SD servers get very Busy as midnight moves across the USA.  Maybe have a build exclusion time from
#   11pm EST for 6 hrs?  Can still 'force' a build from UI SaveSettings if user wants it.
#   could/should disable dns until delayed build occurs
#   BuildExclude = { 'Enable'=True, 'Start'='time', 'Hours'=6, 'DisableDNS'=True}
#   With SD GET/POST retries, I dont think this is necessary, as PiGS handles errors ok
#   User can also control this my just doing a Manual Guide Refresh in the Morning.
#
# V1.40 2024-Jan-12
#  sv.py Typo /mycons
#  sd.py Some cleanup
#  cm.py Change CM server message to .debug - URL confusing when seen in the logs
#  pigsdns.py validate IP addresses in PigsDns 'hosts', with log messages to help
#  Validation of config Server_IP and -s cmd line with help messages
#  Change pigscfg[] to Settings[] in sv.py and ls.py and PigsSettings[] in pigs.py
#  find.htm and se.py - add a checkbox for [x] Start search 'Now' (start near current time)
#  Add timestamp to /pigstatus and save in a file pigstatus.json on every view for debug
#  Updated format of /pigstatus. Move boring stuff to bottom
#  Add UI checkbox UseMyIcons, allows user to disable 'myicons' to check for new SD icons
#    You can always see what Icons are in use with the Lineups(view details) line on the UI
#    when [ ] Use MyIcons is not checked, icons in myicons are bypassed
#  Added file pigsreq.txt, a pip 'requirements' file to install Pigs dependancies
#    After unzipping PiGS files to a folder, open a terminal or cmd window and type
#     sudo pip3 install -r pigsreq.txt   (if pip3 doesnt work, just 'pip' might)
#    This will install/verify python-dateutil, pillow(PIL), and requests in 1 step
#  Added Required modules test and install instructions in case modules are not installed.
#  Added Update Status "Barker" to Channel guide.
#    PigSettings['BarkerChan'] = "3.4"     use = "" to disable
#    Every listing Title on that channel  will contain last
#        DVR Guide Update time like: "Updated Tue 02-13 11:13"
#    Every Listing description will contain last SD Download & Dvr guide update:
#     "Schedule downloaded from SD on Tue 02-13 08:13. This DVR Last Updated Tue 02-13 11:15."
#    I put mine on my QVC channel, which happens to be my 4th channel, so its easy to see
#    Note: SD build time wont (cant) show until DVR does its daily update
#  Improved Startup
#    Service list now Saved and Loaded (i think that is now whole PiGS state)
#    PiGS can now serve DVRs w/o contacting SD when it (re)boots using saved info
#    DNS is turned on immediately if schedule was loaded from disk
#    New command line options -holdoff & -delay <Hrs> can be used
#      to delay the build and/or holdoff DNS (allows sync of DVRs also)
#  Added command line option -delay -d <hours> to Set initial build delay to
#    allow user to set first build for tomorrow morning for example.
#    Also prevents 1st dvr request after boot from resetting build time
#  Added command line option -holdoff(-o) to hold off DNS (thereby DVRs)
#    until after first build occurs. I like my DVRs to update sometime
#    in the morning, but they slip forward over time.
#    I can quit PiGS then restart with
#      sudo python3 pigs.py -holdoff -delay 30 (30hr = 1 day + 6 hrs)
#    This starts PiGS, keeps DNS disabled and schedules the first build
#      in 30 hours (set hours to when you want build to occur)
#    The delay is set to be anytime after your last DVR is expected to update.
#  Added command line option -update -u
#    updates pigs.json with any new configuration keys and exits.  The
#    update happens regardless, this just exits so you can edit immediately.
#  Added Profiling Timers to build for /schedule /programs to see SD timeouts
#  ls.py 1-12-24 Fixed Season/Episode numbers after TVmaze data added 
#  pigs.py fix SaveSettings with -u option, remove Barker default 3.4
#  pigs.service Added KillSignal=SIGINT so PiGS exits/restarts gracefully on 
#    sudo systemctl stop pigs or restart pigs
#  Added dirty flag to UpgradeConfigFile(), user settings now saved if updated
#
# https://stackoverflow.com/questions/11029717/how-do-i-disable-log-messages-from-the-requests-library
# logging.getLogger("requests").setLevel(logging.WARNING)
# logging.getLogger("urllib3").setLevel(logging.WARNING)
#
# V1.39 2023-Jan-04
#  Remove import auto() from enum... its not Python 3.5 compliant and I dont need it.
#  ls.py fix missing Prog detail error description - used pidstr instead of apidstr
#  Add retries on SD GET requests.  Helps build succeed when SD server under heavy load.
#  The Acct API's (PUT/DEL's) do not have retries so if the SD server sends 500 errors
#    a user could skip account verification to maybe get past startup/init
#  Added config key and cmd line -noverify -v to skip SD account verify step on init
#    "VerifyLineups": True = Verify Lineups, False=skip verification
#  Add SD.py msg 'Retry succeeded.  Continuing..."
#  Set build status when loading Sched from disk
#  Added 'CName' functionality to PiGS DNS server for NTP pool redirection
#    e.g. in pigs.json, redirect default pool.ntp.org to server farm of your choice
#   'pool.ntp.org' : 'time.nist.gov',     # redirect time server requests
#  Fix build in LoadProgramsFromDisk - tried to access non-existant schedule
#  Add try/except to utGetPigsVersion() float() conversion in case bad valve
#  Add valid json to GetToken return value to prevent an exception
#  Increase the SD Login timeout from 10s to 90s since SDxx now does retries
#  Added msg "Login Status : Trying..." to UI during login attempt
#  Additional error checking on POST command - validate json data &
#    reject unsupported POST comands (9/21/22)
#  Add GET /linearschedule parameter checking if DVR makes bad startdate
#     requests and return HTTP 400 for bad data (12/28/22)
#  Added PigSettings['SleepMonitor'] key and code for Installs where host PC may sleep
#    Python timers loose time while PC sleeps. Now monitor the expected build
#    Real time and start a build if the time passes (like when the PC wakes up)
#
#  Improved Schedule Remapping support.  DVR+ doesnt like it if it has 
#   duplicate stationID's and will only display one of the schedules.
#   It would show up re-mapped, but maybe gone from previous channel.
#   To get around this, I added a 'schedule' key where you put the
#   CallSign of the stations schedule you want to display.  I use this
#   info to look up the real schedule and this provide a unique StationID
#   for each station so DVR will display the same schedule on both chans
#   You can map one stations schedule to another by adding 'schedule' key.
#   since its json, you can even add a 'comment' key if you wish.
#    {
#      "channel": "15.2",
#      "name": "WMTVDT2 (WMTV-DT2)",
#      "callsign": "WMTVDT2",
#      "affiliate": "CW",
#      "schedule": "WISCDT",
#      "comment": "put channel 3.1 WISCDT schedule on channel 15.2",
#      "favorite": true
#     },
#
#  Added psuedo CallSign MYICON which allows you create a 'fake'
#    station to add icons to stations that do not have SD schedules,
#    or if you want an icon, but want the PSIP OTA schedule.
#    Place the icon file you want in the folder /myicons e.g. /myicons/3.4.png
#    {
#      "channel": "3.4",
#      "name": "WISCDT4 (WISC-DT4)",
#      "realcallsign": "WISCDT4",     <<<<<< changed key of old sign
#      "callsign": "MYICON",        <<<<<<< 
#      "affiliate": "QVC",
#      "favorite": true
#    },
#
# V1.38 2022-May-14
#  Program Schedule and Cache Saved to disk on good build in xxx.pkl
#    Needed way to keep Cache between runs while making code changes 
#    Also ensures a Schedule if there's an SD outage on PiGS boot
#    Added key "SaveSchedule" : 2,
#       0=Dont save, 1=Save Cache, 2=Save Both Schedule & Cache
#    Saved files are loaded if they exist.  Can be manually deleted w/o harm
#    Cmd line flag -n -noload prevents Both from loading
#  Make sure md5 Stale Cache items are in ToSaveSet() so Cache current
#  DNS was left disabled after a failed build (during 20min retry) even
#    if a sched was available (yesterday or loaded)
#    Now DNS is enabled after a build if a schedule is available
#  Made Test_Webserver work when port<>80, allows multiple PiGS instances (for dev)
#  Build days displays (-1/+17 days) which means 1 prior day, 17 fwd counting today
#
#
#
#---------------------------------------------------------------------------------------
# ToDo list
#
# scan code for ToDo's and fix or remove
#
# Could this code base be used with another guide provider?
# Probably.  The key tasks would be to :
#     1) Build the list of channel services BuildServicesListNEW() using their data
#     2) Build the Linear schedule list BuildLinearSchedule() using their data
#     3) Plus a lot of UI stuff, like out how to login, configure the lineups(UI changes)
# The sequencing in this app was built based on the api provided by Schedules Direct
# If you were able to emulate the functions in SD.py (ie reformat everything there)
# using another providers data to provide equivalent results, it might go pretty 'smoothly'.      
#---------------------------------------------------------------------------------------

PIGS_LOG_FILE     = "pigslog.txt"
MAX_LOGFILE_BYTES = 250000  # 150K then rotate log file to next
LOGFILE_BACKUPS   = 3       # Current file and 3 backups

#---------------------------------------------------------------------------------------
# PiGS can act as a DNS server.  The HostsDict definition serves as a 'hosts' file
# Pigs dns will translate names found here to an IP.
# Any request that does not match will be resolved with gethostbyname()
#
# The process works as follows:
#   1) Check for a CName substitution.
#      If a name match is found in HostDict then
#         if its not '*' or an IP (4 digits with 3 dots) do substitution
#   2) This HostDict is checked, if a match is found then
#      return the specified IP or the PiGS IP (for *) and exit
#   3) If no HostDict match, then request sent to gethostbyname() which
#     2a) checks for a match in _etc_hosts,
#        If found return the IP, else
#     2b) Use the configured primary dns server (probaby router gateway)
#   The results are returned to DVR
# V1.39 supports cname type fuctionality. e.g. dvr always does NTP to pool.ntp.org
#   You can get it to use time.nist.gov as follows
#   'pool.ntp.org': 'time.nist.gov',
#   another use of cname, if the DNS record for any of the 4 'online' URLs were removed
#   DVR could be fooled by using CNAME
#   'ecostar.com':'facebook.com',   

HostsDict = {                               # PiGS DNS will replace '*' with PiGS discovered IP or Server_IP 
    'epg.channelmastertv.com' : '*',        # fw 1.35 guide requests use this URL
    'roproxy-guide.echostarcdn.com' : '*',  # fw 1.34 and before guide requests use this URL
    "stb.channelmastertv.com": "0.0.0.0",   # Blocks requests to dead Channel Master TV service
                                            # Bock firmware updates 
    'tr50.dishaccess.tv': '0.0.0.0',        #   tr50.dishaccess.tv CNAME stbfw.echodata.tv A 67.148.153.225
    'stbfw.echodata.tv' : '0.0.0.0',        #   firmware update url
    
    # These are here for documentation.
    # The leading '_' prevents a name match so requests are forwarded to DNS.
    # Any of these could be redirected if needed
    
    '_cps-static.rovicorp.com' : '*',       # Rovi Icons from here - do not block with 0.0.0.0
    '_pool.ntp.org' : 'forwarded url',      # NTP server requests,  forwarded
    '__pool.ntp.org': 'time.nist.gov',      # A way to redirect NTP requests to another server by cname
    
                                            # DVR+ uses these next 4 to determine if it's Internet connected
    '_microsoft.com': 'forwarded url',      # with the leading '_' these are forwarded to a real DNS server.
    '_echostar.com' : 'facebook.com',       # If global DNS record for E* was removed, we can make it work.
    '_amazon.com'   : 'forwarded url',      # DVR+ doesnt like it if I return IP's on the same network
    '_google.com'   : 'forwarded url'       # DVR+ doesnt access these addresses, it just wants a DNS response.
}


#---------------------------------------------------------------------------------------
# Json format configuration file

CONFIG_FILE = 'pigs.json'

# Config json file is read into this global config dictionary at boot and on UI Save

PigSettings = None    # global Settings dictionary, empty until pigs.json is loaded

httpd = None  # webserver handle
  
# This is the layout of the pigs.json master config file

# IF NEW KEYS ARE ADDED TO THE DICTIONARY, THEN ADD AN INIT LINE IN UpgradeConfigFile() below 

# key types System(s), UserInterface(UI), User(u), System & Command line(sc)

PigsConfigDict = {
                          # Required keys
  "sdusername"   : "",    #s your SD username
  "sdsha1pwd"    : "",    #s your SD password, sha1 encoded
  "sdApiVersion" : "",    #s Allows for testing different SD API's
  "Server_IP"    : "0.0.0.0", #s 0.0.0.0 means pigs will monitor port 80 on all IP's (wire & wired interfaces)
  "ThreadedHTTP" : False, #s Use threaded HTTP.server.  not sure my code is thread safe.  DONT USE
  "IDE_Port"     : 0,     #s http port over-ride for development, typ 8000.  0 means do not override
  "RebuildSecs"  : 82800, #s 82800 = 23 hours = 23h*60*60 Next Build Time in seconds
                          # If RebuildSecs is set to 0, Pigs Should only Auto-Build at same time each day.
  "StaticBuildTime": "",  # Build hour and minute in 24 Hr time  07:00 or 3:30pm = 15:30
                          # was FixedBuildTime in 1.35/1.36 (had a 07:00 build time by mistake)
                          #
  "SleepMonitor" : True,  #s Background now Monitors Expected Build Time in case PC was sleeping
  "WakeNotify"   : True,  #s Display a message when OS has caused pigs to sleep
  "FirstBuildDelay": 30,  #s First build occurs 30s after boot to allow platform to stabilize. 0 disables build
  "BuildModeDelete": False, #s False= Update Mode - Keep last build until todays is sucessful.
                            #  True = Delete sched before build(saves memory for PiZero?)
  "VerifyLineups": True,  #sc Init step to verify lineups, also cmd line override 1.39
  "NetworkWait"  : 60,    #sc Wait for up to 60s for Network to become ready
  "AutoSaveConfig": 2,    #s  Save config file every N Builds, 0 will disable auto-save

  "BarkerChan"   : "",    #u  eg "3.4", Pigs will set Show title to build & update times in every slot
  "IconType"     : 1,     #ui 0=No Icons, 1=Sched Direct, 2=CM/Rovi Icons, 3=CutePigs???
  "CacheCMIcons" : True,  # keeps a local copy of the CM / Rovi Icons
  "CacheSDIcons" : True,  # keeps a local copy of the resized SD Icons
  
  "SaveFsSDIcons": False, # True Saves a copy of the FullSize SD icons in ./sdicons for external manipulation
  "SDImSize"     : [72,54], # 360w x 270 / 5 = 72 x 54 (json cant do Tuple, so will use a list)
  "SDImResample" : 3,     # Default to type 3.   Pillow Lib .resize() Resample type 0..5
  "SDImReduceGap": None,  # Pillow 7.0 Reducing Gap 
  "SdIconTheme"  : "dark", # choices are dark, gray, light, white
  
  "UpdateIconCache" : False, #ui set true when icon caches need to be updated
  "UseMyIcons"   : True,  # Default True, check /myicon folder for custom icons
  "DefaultIcon"  : "pinkpig.png",  # if no guide icon is found... another could be "92x36box.png"   
  "PrependEpisode" : True,  # Prepend Episode Title, SeasonEpisode #'s, FirstAirDate
  "AppendCast"   : True,  # Append cast info to description

  "LoggingLevel" : 20,                # default logging level. DEBUG 10,INFO 20,WARNING 30, ERROR 40,CRITICAL 50
  "LoggingFile"  : PIGS_LOG_FILE,     # filename of log file, like pigslog.txt
  "LogFileBytes" : MAX_LOGFILE_BYTES, # 150K then rotate log file to next
  "LogFileCount" : LOGFILE_BACKUPS,   # Current file and 3 backups
  
  "AlertEmails"  : "",    #u csv emails to receive alerts "Name1<email1@abc.com>,Name2<email2@xyz.com"
  "SmtpUser"     : "",    #u gmail address  "someaddr@gmail.com"
  "SmtpPass"     : "",    #u gmail password, make a new account just for pigs

  "PiGS_Version" : PIGS_VERSION, #s Version string so UI index.htm can display
  "PiGS_NewVer"  : 0.0,   #ro Read-Only.  Ver from pigs.goihl.com/zips/version.txt for UI update msg 1.36
  "CheckVersion" : True,  #s  check pigs.goihl.com/zips/version.txt for new version of PiGS
  "SaveSchedule" : 2,     #s  0=Dont save, 1=Save Cache, 2=Save Program Schedule & Cache 
  "PiGS_IP"      : '0.0.0.0', #ro Read-Only. IP Addr string so UI index.htm can display
  "PinkPigPos"   : 0,     #s display 1 pink pig icon at pos N on the guide or on Chan x.y, use 0 to disable
  "AllowAnyZip"  : True,  # allow any text in Zip box.  False = USA 5 digit, numbers only
                          #    Right now US and Canadian Work - Have not tried Mexico
  "PigsDoesDNS"  : True,  #ui PiGS acts as a DNS Server for DVR+.  Forward requests for msft,google,amzn, etc
  "PigsDelayDNS" : True,  #s  PiGS starts DNS server thread after Schedule is built  
                          # Development keys
  "ChunksEnabled": False, # Enable Http/1.1 chunks for schedule send
  "MGRnewSched"  : True,  # Manual Guide Refresh forces Rebuild of SD Schedule
  "debugToken"   : "",    # If defined, over-rides the SD token request (allows concurent use with PostMan)
                          #   Generate the token in PostMan and enter it here
  "DaysLimit"    : 0,     # Num days fetched from SD (DVR+ can only use 15), 0 for Max Days, less than 17 yields 403 errors
                          # DaysLimit was used to limit day to 2 or 3 for faster builds during development.
  "DisplayDVR"   : 7,     # pigstatus display dvrs updated within last N days, 0 for all 
  "LogHttpReq"   : False, # debug enable annoying http logs on each send
  "SaveJson"     : False, # save the PiGS Channel Lineup sent to DVR+ in DvrLineup.json for troubleshooting
  "TimeZones"    : [      # This gets copied into the DVR+ message response headers
      {
        "StartDateTime" : "2020-03-08T08:00:00Z",
        "EndDateTime"   : "2020-11-01T07:00:00Z",
        "Offset" : "???"  # ??? will be initialized with users local as found in their system
      }
    ],
  "DnsHosts" : HostsDict, # DNS Host dictionary for PiGS DNS server (functions like /etc/hosts )
  
  "DVR_States" : {},  # States of dvr's using PiGS stored here 
  
  "MarkAsNew": [   # User
         # [ "Title Contains", "Copy Contains", ForceNotNew], # ForceNotNew Flag Marks all Title matches as NOT New, ex those that match copy 
         # [ "America's Test", "S21", False],       # Example mark all Season 21's of America's Test Kitchen as New
         # [ "NFL Football",   "Packers, True],     # Example Mark NFL Football as NOT New, except Mark the Packers as New
         [ "Show1", "Season S3 or other Copy to match", False],
         [ "Show2", "Season or other Copy to match",    False],
         [ "Show3", "NoComma after Last item in list",  False]
      ],
                          # UI Finally, list of users lineups, limited to 4 by Schedules Direct
  "MyLineups"    : []     # List of LineupDef objs with users favorite stations flagged
                          # Multiple Lineups allow user to pull schedule data from far away stations.
                          # e.g. their local zipcode might not include all the stations they can receive.
                          # MyLineups will be used to build the ServiceDetails station List DVR+ requests  
}


# User Settings stored in usersettings.json or mysettings or mypigs.json????

# in Dev for 1.40
# User Lineups are processed first and supercede the SD channel lineup
# This wont work b/c the channels have to be associated with an SD lineup
UserLineup = {   # LineupDef
  "LineupName" : "UserLineup",   # SD Name for the lineup
  "ChannelList": []              # List of UserChannelItems.
  }

# What is this???
#  Added psuedo CallSign MYICON which allows you create a icon only station
#    to add icons to stations that do not have SD schedules,
#    or if you want an icon, but want the PSIP OTA schedule.
#    Place the icon file you want in the folder /myicons e.g. /myicons/3.4.png
#    {
#      "callsign": "MYICON",        <<<<<<< 
#      "schedule": "WISCDT",


UserChannelItem = {
    "channel": "3.4",
    "name": "WISCDT4 (WISC-DT4)",
    "callsign": "WISCDT4",
    "affiliate": "QVC",
    "favorite": True
  }

def UpgradeConfigFile() :   # Auto-Upgrade pigs.json
    
    # flag to tell if any setting has been updated so we can save new settings
    dirty = False   

    if 'UseMyIcons' not in PigSettings :        # 1.40 keys
        dirty = True
        PigSettings["UseMyIcons"] = True
        
    if 'BarkerChan' not in PigSettings :
        dirty = True
        PigSettings["BarkerChan"] = ""

    if 'SleepMonitor' not in PigSettings :      # 1.39 keys
        dirty = True
        PigSettings["SleepMonitor"] = True

    if 'WakeNotify' not in PigSettings :   
        dirty = True
        PigSettings["WakeNotify"] = True

    if 'VerifyLineups' not in PigSettings :     # Init step to verify lineups, also cmd line override 
        dirty = True
        PigSettings["VerifyLineups"] = True

    if 'SaveSchedule' not in PigSettings :      # 1.38 keys
        dirty = True
        PigSettings["SaveSchedule"] = 2         # pickle sched and cache to disk 

    if 'LogFileBytes' not in PigSettings :      # 1.37 keys
        dirty = True
        PigSettings["LogFileBytes"] = MAX_LOGFILE_BYTES  # 250K then rotate log file to next

    if 'LogFileCount' not in PigSettings :      # 1.37 keys
        dirty = True
        PigSettings["LogFileCount"] = LOGFILE_BACKUPS  # Current file and 3 backups

    if 'ChanFilter' in PigSettings :           # Remove unused key
        dirty = True
        PigSettings.pop("ChanFilter",0)

        # version from http://pigs.goihl.com/zips/version.txt to print update msg 1.36
    if 'PiGS_NewVer' not in PigSettings :      # 1.36 keys
        dirty = True
        PigSettings["PiGS_NewVer"] = 0.0
    if 'CheckVersion' not in PigSettings :      
        dirty = True
        PigSettings["CheckVersion"] = True

    if 'LastSvcBuild' in PigSettings :         # Remove unused keys in 1.35
        dirty = True
        PigSettings.pop("LastSvcBuild",0)
    if 'LastSchdBuild' in PigSettings :        # Remove unused key
        dirty = True
        PigSettings.pop("LastSchdBuild",0)

    if 'ThreadedHTTP' not in PigSettings :     # 1.30 Use threaded HTTP.server. not sure all code is thread safe
        dirty = True
        PigSettings["ThreadedHTTP"] = False  

    if 'StaticBuildTime' not in PigSettings :  # 1.30/1.37 Clock scheduled build time
        dirty = True
        PigSettings["StaticBuildTime"] = ""      # Build H,Min 24 Hr time  7:00AM "07:00"  3:30pm = "15:30"
    # in 1.30-1.36, I had an accidental default time of '07:00' in 'FixedBuildTime' delete so DVR auto schedules
    if 'FixedBuildTime' in PigSettings :
        dirty = True
        PigSettings.pop("FixedBuildTime",0)      #  Remove FixedBuildTime key, now its called StaticBuildTime
     

    if 'sdApiVersion' not in PigSettings :     # 1.24 Allows for testing different AD API's
        dirty = True
        PigSettings["sdApiVersion"] = ""         # Empty means no version over-ride
    if 'BuildModeDelete' not in PigSettings :  # 1.24 
        dirty = True
        PigSettings["BuildModeDelete"] = False   # True = delete current schedule before starting build

    if 'AlertEmails' not in PigSettings :      # 1.24 CSV email list to receive alerts 
        dirty = True
        PigSettings["AlertEmails"] = ""          # 
    if 'SmtpUser' not in PigSettings :         # 1.24 smtp email/gmail address  "someaddr@gmail.com"
        dirty = True
        PigSettings["SmtpUser"] = ""             # 
    if 'SmtpPass' not in PigSettings :         #  1.24 smtp email/gmail password, make a new account just for pigs
        dirty = True
        PigSettings["SmtpPass"] = ""             # 

    if 'AutoSaveConfig' not in PigSettings :   # 1.24 Save config file ever N days
        dirty = True
        PigSettings["AutoSaveConfig"] = 2    # 

    if 'MarkAsNew' not in PigSettings :        # 1.24 mark as new shows
        dirty = True
        PigSettings["MarkAsNew"] = []    #
        PigSettings["MarkAsNew"].append( ["Show1", "Season or other Copy to match", False] )
        PigSettings["MarkAsNew"].append( ["Show2", "Season or other Copy to match", False] )
        PigSettings["MarkAsNew"].append( ["Show3", "NoComma after Last item in list", False] )
      
    if 'NetworkWait' not in PigSettings :   # First build delay is 30s after reboot/start 
        dirty = True
        PigSettings["NetworkWait"] = 60   # Wait for up to 60s for Network to become ready

    if 'UpdateIconCache' not in PigSettings :   # 1.24 icon cache update when changed
        dirty = True
        PigSettings["UpdateIconCache"] = False    # 

    if 'FirstBuildDelay' not in PigSettings :   # First build delay is 30s after reboot/start 
        dirty = True
        PigSettings["FirstBuildDelay"] = 2   # Upgrades get 2s so as not to mess with existing installs

    if 'IDE_Port' not in PigSettings :     # change default http port for working in IDE
        PigSettings["IDE_Port"] = 0        # Default new key to 0
    if 'Server_Port' in PigSettings :      # Remove unused key
        PigSettings.pop("Server_Port",0)
        
    if 'ChunksEnabled' not in PigSettings :     # Enable Http/1.1 chunks for schedule send
        PigSettings["ChunksEnabled"] = False    # Default new key to false, Chunking requires HTTP/1.1

    if 'DnsHosts' not in PigSettings :        # PiGS acts as a DNS Server for DVR+.  Forwards some requests.
        dirty = True
        PigSettings["DnsHosts"] = HostsDict   # Add HostsDict if it does not exist
    if 'stb.channelmastertv.com' not in PigSettings['DnsHosts']:  # add host I forgot
        dirty = True
        PigSettings['DnsHosts']["stb.channelmastertv.com"] = "0.0.0.0"

    if 'DisplayDVR' not in PigSettings :      # pigstatus display dvrs updated within last N days, 0 for all
        dirty = True
        PigSettings["DisplayDVR"] = 7

    if 'IconType' not in PigSettings :        # Icon type 1=SD, 2=CM
        dirty = True
        PigSettings["IconType"] = 2 if PigSettings["UseCMImages"] else 1
        PigSettings.pop('UseCMImages',0)      # remove old key nolonger used default 0 prevent exception
        
    if 'PigsDoesDNS' not in PigSettings :     # PiGS acts as a DNS Server for DVR+.  Forwards some requests.
        dirty = True
        PigSettings["PigsDoesDNS"] = False    # Default new key to false to not mess up current users
    if 'PigsDelayDNS' not in PigSettings :    # PiGS acts as a DNS Server for DVR+.  Forwards some requests.
        dirty = True
        PigSettings["PigsDelayDNS"] = True    # Default new key to True, start DNS after Schedule is built
    if 'Server_IP' not in PigSettings :
        dirty = True
        PigSettings["Server_IP"] = '0.0.0.0'  # defalt "0.0.0.0"

    if 'RebuildSecs' not in PigSettings :
        PigSettings["RebuildSecs"] = 82800  # 23 hours = 23h*60*60 Next Build Time in seconds
    if 'MGRnewSched' not in PigSettings :
        PigSettings["MGRnewSched"] = True    # Manual Guide Refresh forces Rebuild of SD Schedule
    if 'AllowAnyZip' not in PigSettings :
        PigSettings["AllowAnyZip"] = True    # allow any text in Zip box.  False = USA 5 digit, numbers only
    if 'SaveFsSDIcons' not in PigSettings :
        PigSettings["SaveFsSDIcons"] = False # keeps a local copy of the resized SD Icons
    if 'SDImSize' not in PigSettings :
        PigSettings["SDImSize"] = [72,54]    # 360w x 270 / 5   = 72 x 54
    if 'SDImResample' not in PigSettings :
        PigSettings["SDImResample"] = 3      # Deault to BICUBIC.   Pillow Lib .resize() Resample type 0..5
    if 'SDImReduceGap' not in PigSettings :
        PigSettings["SDImReduceGap"] = None  # Only in Pillow 7.0 - Reducing Gap
    if 'CacheSDIcons' not in PigSettings :
        PigSettings["CacheSDIcons"] = True  # keeps a local copy of the resized SD Icons
    if 'CacheCMIcons' not in PigSettings :
        PigSettings["CacheCMIcons"] = True   # keeps a local copy of the CM / Rovi Icons
    if 'DefaultIcon' not in PigSettings :
        PigSettings["DefaultIcon"] = "pinkpig.png"
    if 'SdIconTheme' not in PigSettings :
        PigSettings["SdIconTheme"] = "dark" # choices are dark, gray, light, white
    if 'TimeZones' not in PigSettings :
        PigSettings['TimeZones'] = copy.deepcopy(PigsConfigDict['TimeZones'])
            
    if PigSettings['TimeZones'][0]['Offset'] == "???" :  # ??? is flag to tell us to init
        # Set the timezone object to the current time zone on each boot
        # The TimeZones do nothing for pigs/dvr.  dvr gets zone info from NTP or PSIP
        # Makes sense b/c dvr is meant to work offline PSIP only
        s,offset = LocalTimeStr( datetime.utcnow() )
        PigSettings['TimeZones'][0]['Offset'] = str(offset)

    return dirty

#---------------------------------------------------------------------------------------
# UI Settings of config
# A full copy of the PigsConfigDict is sent to the UI index.htm for manipulation
# Some is for display only, and some is for getting the users settings.
#
# The following PigsPreserveSettings dict is a list of keys that will be preserved
# when the user hits Save Settings.
#
# Anything in this list becomes essentially READ-ONLY - UI cant change it.
# E.g, if a guide update occurred while the user had the UI open and the
# LastSchdBuild time had been updated, we want to preserve that key value since
# it occurred after the UI was opened.
# Another example is the sha1 password.  Initally it does not exist in the config file.
# User presses verify login - UN & PW is sent to the server and the  sha1 is created
# and saved in the server conig, but not sent back to the UI so the UI data is 'out of date,'
# so the server side sha1 is 'Preserved' when the user clicks save settings.
# The key names here must match those in PigsConfigDict exactly
# The values in PigsPreserveSettings do nothing - Just place holders.
# I suppose I could just make a straight List [] object of key names instead of a dict object.

PigsPreserveSettings = {
  "sdsha1pwd"    : "",  # your SD password, sha1 encoded
  "PiGS_Version" : "",  # Version string for loggin
  "PiGS_IP"      : "",  # IP string for loggin
  "debugToken"   : "",  # this key over-rides the token request
  "DaysLimit"    : "",  # limit number of schedule days fetched from SD
  "LogHttpReq"   : "",  # debug enable annoying http logs on each send
  "SaveJson"     : ""   # save DVR+ json request body
}

# A lineupDef is created when the User clicks 'Preview Lineup' with a valid ZIP
# 
# It will be created by PiGS and sent to the UI but exist only on the UI side until the user
# clicks 'Add This Channel Lineup' (added to MyLineups[]) followed by Save Settings (send to PiGS server)
# 'Add..." action will add the 'Preview" to the UI copy of config MyLineups[].
# So the MyLineups array can/will change size from what was sent to the UI
# 
# To commit UI changes to the Server, they must click 'Save Settings' to send to PiGS
#
# SD allows for up to 4 lineups
# LineupDef dict is one user lineup: the LineupName for UI display, and list of favorite chans.
# MyLineups[] is a list of LineupDef's in pigs.json

LineupDef = {
  "LineupName" : "LU-Name",   # SD Name for the lineup
  "ChannelList": []           # List of ChannelDef objs. It will include every channel in the lineup
  }

# The ChannelDef object below is for reference only.  The first 4 keys come from SD preview call.
# The SD preview will send the channel,name,callsign,and affiliate keys, I will add the favorite key
# after getting the results from the SD command.

# When the user saves the lineup, the DVR+ services list will be created using the
# channels marked as favorites and info obtained from SD station Map command

# These are the key names sent from the SD lineup preview results, with the addition of
# my favorites key.  We will loop through and add the favorite key after getting the data
# from SD so the favorite key exists for the UI and has an initial value
# There are a couple special use keys and cases to get 

ChannelDef = {
    "channel"  : "3.1",    # with example values.
    "name"     : "WISCDT (WISC-DT)",
    "callsign" : "WISCDT",
    "affiliate": "CBS",
    "favorite" :  True      # by default.
}

#---------------------------------------------------------------------------------------
# References for HttpServer object
#https://github.com/python/cpython/blob/3.8/Lib/http/server.py

from http.server import BaseHTTPRequestHandler, SimpleHTTPRequestHandler

from io import BytesIO
from urllib.parse import parse_qs, urlparse

# CM/DVR+ bases channel lineups on Service ID's  Multiple ZIP's might map to one service ID
# PiGS will be configured up front using Schedule Direct lineups 
# Going to use a 'dummy' service ID of 99999 everywhere.
# Every DVR+ zip request will get a service id of 99999
# The 99999 number is important b/c it must be consistent across all DVR+ service calls,
# listing calls, and program airings or DVR+ wont display it.
# The actual channel lineups will be configured in the UI / via Schedules Direct

PiGS_SERVICEID = 99999

#running starting a program at bootup on a Pi
#https://www.dexterindustries.com/howto/run-a-program-on-your-raspberry-pi-at-startup/

#https://docs.python.org/3.7/library/logging.html?highlight=logging#module-logging
#https://www.blog.pythonlibrary.org/2014/02/11/python-how-to-create-rotating-logs/

import logging
import logging.handlers
import time
from   datetime import datetime, timedelta
from   dateutil import parser
import json
import gzip, zlib  # jsgz = gzip.compress(js), zlib does stream
import sched
from   copy import copy
import copy

# Functions to send data to DVR+
from cm import CMInitAltServer
from sd import GetSha1Password, SDGetToken, SDGetLineupPreview, SDAddDeleteLineups, SDSetAPIVersion
from sv import GetServiceIdFromZip, GetServiceDetails, BuildServicesListNEW, GetServiceList, GetSVCBuildInfo
from ls import BuildLinearSchedule, GetLinearSchedule, GetChunkedLinearSchedule
#LSClearSchedule, 
from ls import LsGetDVRStatus, LS_GetDictOfDVRs, LS_SetDictOfDVRs
from ls import LSGetProgramSchedule, lsr, LSBuildStat
from se import SearchSchedule, DisplayLineup

from pigsdns import PigsDnsStart, PigsDnsStop, PigsDnsAlive, PigsDnsStatus

# utility functions
from ut import LocalTimeStr
from ut import VersionStringToFloat
from ut import SendEmailAlert
from ut import EM_Initialize
#from ut import test_serverport
from ut import test_WebServer, utGetPigsVersion

bLogHttp = False  # a flag to stop some annoyingly large httplib debug messages

#https://docs.python.org/2/library/os.path.html
import shutil
from os import path
import os
import signal

#=========================================================================
# Save config file, keep a history of 5 backups
# https://docs.python.org/3/library/shutil.html
# https://www.w3schools.com/python/python_file_remove.asp
# https://docs.python.org/3/library/shutil.html
# https://docs.python.org/3/library/os.path.html
# https://docs.python.org/3/library/os.html#module-os
# I wonder if there is some type of Rotation Save function in Python

CFG_HISTORY = 5

def SaveSettingsFile(configDict, configFile=CONFIG_FILE) :
    
    # separate file path/name from its extention
    root, ext = path.splitext(configFile)

    # delete the oldest log file #5
    dst = root+str(CFG_HISTORY)+ext
    if os.path.exists(dst):
        os.remove(dst)
    
    # then rename files to keep history
    # ren 4 to 5, 3 to 4, 2 to 3, etc
    for i in range(CFG_HISTORY-1,-1,-1) : # start, stop 1 before, step
        src = root+str(i)+ext if (i>0) else root+ext  # strip '0' for current file 
        dst = root+str(i+1)+ext # to 5
        if os.path.exists(src):
            os.rename(src, dst)  
        #src = path.realpath("guru99.txt");
        #os.replace(src, dst, *, src_dir_fd=None, dst_dir_fd=None)

    # and write config file to storage 
    # pretty json with indent=2.  sorted is ugly and ungrouped
    with open(configFile, "w") as outfile:    
        json.dump(configDict, outfile, sort_keys=False, indent=2) 
    
    outfile.close()

    #TodDo  # seems to be only 644 so user cant delete file. do i want that???

#---------------------------------------------------------------------------------------
    
def CheckForPigsUpdate(ModName=None) :
    
    global PigSettings    
    pver = 0.0
    msg  = ""

    # get current ver from pigs.goihl.com/zips/version.txt
    if PigSettings["CheckVersion"] :
        logging.info("%s Checking for a new version of PiGS...", ModName)
        pver = utGetPigsVersion()  

    PigSettings["PiGS_NewVer"] = pver  # save in config file for UI Alert Msg
    
    # Strip version from string "V1.35 2022...."
    if pver > float(PigSettings["PiGS_Version"][1:6]) : 
        
        msg = 'There is a new version(%s) of PiGS at http://pigs.goihl.com' % (pver)
        logging.warning("*****************************************************************")
        logging.warning("* " + msg + " *")
        logging.warning("*****************************************************************")
    elif pver < 0.0 :
        msg = 'PiGS Version check error Version(%s)' % (pver)
        logging.warning("*****************************************************************")
        logging.warning("* " + msg + " *")
        logging.warning("*****************************************************************")
        
    return msg

#---------------------------------------------------------------------------------------
# The schedule build can take a 'long' time, a minutue or two (or even more).
# To allow the web server to respond to multiple DVR requests, UI commands, and
# web pigstatus requests in real-time, the scheudle build runs in a separate thread.  
# If a DVR request comes in while the Build is in progress that cant be answered,
# the webserver will report a http 503 server busy error.
# This will allow the webserver to respond to all requests (with a 400 or 503 at least)
# instead of ignoring requests when the build is in progess.
# It appears requests were stacking up, timing out, confusing the http.server resulting in
# low level "broken pipe" errors.
#
# A Timer thread is used to initiate the build about 1 hour before a DVR typically requests it.


import queue
import threading

BuilderThread   = None               # Thread Handle
BuilderQueue    = queue.Queue()      # Queue to send Buildxx commands to Builder
BuildAbortEvent = threading.Event()  # Event used to abort Build if needed.  isSet()  set()  clear()  wait(timeout)
SDResourceLock  = threading.Lock()   # Schedules Direct Resource Lock, only 1 command can use SD at a time

WebserverResetEvent = threading.Event() 
WebserverResetEvent.clear()

# These commands access SD and need SDResourceLock
BuildLogin     = "Login"
BuildServices  = "Services"    
BuildSchedule  = "Schedule"
BuildTerminate = "Terminate"
BuildAddRemove = "AddRemove"
BuildPreview   = "Preview"

# These commands do not need SDResourceLock
BuildStartDNS  = "DnsStart"
BuildWebserver = "WebServer"
BuildDelay     = "Delay"          # BuilderQueue.put( {'Command':BuildDelay, 'Params':10} )
BuildRelease   = "ReleaseLock"
BuildInit      = "InitWOBuild"   # 

# Initialization steps can fail for various reasons - DNS errors,
#   SD/CM server errors, etc.
# Keep track of where we are so we know what is left to initialize.

from enum import IntEnum

class InitState(IntEnum) :
    NOT_READY   = 0    # Not Initialized
    CM_ALT_URL  = 1    # Init the CNAME url for CM
    SD_LOGIN    = 2    # Test SD Login Credentials
    SD_LINEUPS  = 3    # Verify PIGS matches SD Lineups
    SERVICE_RDY = 4    # DVR Service List Built
    SCHEDL_NOTRDY=5    # Schedule has been cleared pending build
    SCHEDL_RDY  = 6    # Schedule has been Built
    # auto() requires Python 3.6
    #INIT_COMPLETE = auto()  # SD verified, Schedule built
    
ReadyState = InitState.NOT_READY

AutoSaveCounter = 0  # config file auto-save counter

def ResetReadyState() :
    global ReadyState
    ReadyState = InitState.NOT_READY

# Commands to the Builder are sent via a dict structure.
#  'Command' key is required.
#  The remaining keys are defined by the command
BldCmd = {
    'Command' : None,    # one of the Above Buildxx commands
    'Params'  : None,    # Params to pass in.  See BuildAddRemove for example
    'Event'   : None,    # If caller needs to be signaled when result ready. See BuildPreview for example
    'Result'  : None     # Results to send back to caller.
                        #    A dict works good.  See BuildPreview for example
  }

# Handles to build timers -
#   DvrBuildTimer is set by first DVR that contacts PiGS after a build
#   WDogBuildTimer is a timer if a DVR doesnt contact PiGS for 24 hrs

WDogBuildTimer  = threading.Timer(1,None)   # watchdog build timer 
DvrBuildTimer   = threading.Timer(1,None)
NextBuildTimeStr= ""
AutoBuildTimeStr= ""
gExpectedBuildTimeUTC = None  # variables 
gLastPollTimeUTC = None

# DoWDogBuildTimer() is a watchdog timer -
# It ensures that the schedule gets built at lease every 24 hrs
# This timer is activated right after the Schedule build completes and
#  and fires if a DVR never contacts pigs to set the 23hr timer
# This timer is also used to perform the Fixed Time 'static' build

def DoWDogBuildTimer():
    
    global gExpectedBuildTimeUTC
    
    gExpectedBuildTimeUTC = None    # reset the Polling build backup time test
    
    if ( PigSettings["StaticBuildTime"] ) :
        logging.warning("THRD:DoWDogBuildTimer: Performing Daily Auto-Build")
    else :
        logging.warning("THRD:DoWDogBuildTimer: has fired. Where are the DVR's?")

    # Trigger the normal Timer build process 0.1 second from now
    # Kicking off the Build timer will also reset this Auto build timer.
    SetDvrBuildTimer(0.1)


# DoTimerBuild is called when the DvrBuildTimer thread fires
# The Timer thread is dead after the Build command is Queued
# The build timer will be reset when a dvr makes a schedule request.
# The Timer build is not scheduled if the username/pwd has not been set (like on first run)
# 

def DoTimerBuild(bDoScheduleBuild):
    global WDogBuildTimer
    global AutoBuildTimeStr
    global ReadyState
    global httpd
    global RestartServerRequest
    global AutoBuildTime
    global port # webserver port for testing
    global gVerifyLineups
    global gExpectedBuildTimeUTC

    gExpectedBuildTimeUTC = None    # reset the Polling build backup time test

    at  = datetime.utcnow() # Display local time the timer Fires
    FiredTimeStr,tz = LocalTimeStr(at, showSeconds=True)
    
    logging.info("THRD:DoTimerBuild: has fired at %s. Waiting for SDResourceLock Lock...", FiredTimeStr)

    DoInit = True

    # Building the Schedule requires the SD 'resource' 
    # Do a hard wait for the SD Lock.  waits until the Build Queue is empty
    if SDResourceLock.acquire(True) :
        
        logging.info("THRD:DoTimerBuild: Acquired SDResourceLock.")
        
        # Todo - SDResourceLock there could be race conditions if commands get processed faster than they are added to queue?

        #---------------------------------------------------------------
        # check for newer PiGS version at pigs.goihl.com
        
        updateMsg = CheckForPigsUpdate("THRD:BUILDER:")

        #---------------------------------------------------------------
        # Verify that PiGS Webserver (typ port 80) is responding to http requests
        # this might break if PiGS started with a port other than 80 before V1.38

        wsAddr = PigSettings['PiGS_IP'] + ":" + str(port)
        
        logging.info("THRD:BUILDER: Testing PiGS Webserver at %s", wsAddr)
        
        if test_WebServer( wsAddr ) == 0 :
            
            logging.info("THRD:DoTimerBuild: WebServer is Responding")
            
        else :
            logging.warning("*******************************************************************")
            logging.warning("****** THRD:DoTimerBuild: PiGS WebServer is not responding!  ******")
            logging.warning("*******************************************************************")
            
            logging.warning("THRD:DoTimerBuild: Queuing WebServer Restart.")
            BuilderQueue.put( {'Command':BuildWebserver} )
            
            # From https://docs.python.org/3/library/socketserver.html#socketserver.TCPServer
            # httpd.shutdown()        # This will like a Ctrl-c to break main loop(doesnt work)
            # Tell the serve_forever() loop to stop and wait until it does. shutdown() must be called
            # while serve_forever() is running in a ***different thread*** otherwise it will deadlock.
            # signal main loop to restart WebServer
            #RestartServerRequest = True
            # Setting private variable __shutdown_request = True will break the serve_forever() loop
            #httpd._BaseServer__shutdown_request = True
            
        #---------------------------------------------------------------
        # Added ReadyState Initialization in V1.22.
        # If Startup Initialization did not complete due to some error,
        # complete any remaining Initialization here.
        # If there is Schedule Build error, a timer is set for 20 minutes
        
        if ReadyState <  InitState.CM_ALT_URL :
            logging.info("THRD:DoTimerBuild:INIT: Setting Default CM Server.")
            cmAltURL   = CMInitAltServer()
            ReadyState = InitState.CM_ALT_URL  # not going to fail if cant get Alt URL
            #if CMInitAltServer() :  # we can do this right here.  
            #    ReadyState = InitState.CM_ALT_URL
        
        #---------------------------------------------------------------
        # Test SD Connection by attempting a GetToken()
        
        if ReadyState <  InitState.SD_LOGIN :
            logging.info("THRD:DoTimerBuild:INIT: Verifying SD connection(GetToken)")
            un = PigSettings['sdusername']     # default: use username and pwd from pigs.json
            pw = PigSettings['sdsha1pwd']
            code, tokencode, errj = SDGetToken( un, pw, PigSettings.get('debugToken'))
            if (code == 200 and tokencode == 0) :  # Token get successful -
                ReadyState = InitState.SD_LOGIN
            else :
                logging.warning("THRD:DoTimerBuild:INIT: SD_LOGIN Error(%d) %s", code+tokencode, errj)
                DoInit = False
                # if we could not get a token for whatever reason, abort the build commands 
                # BuildAbortEvent.set()   #signal Linear Schedule Builder to abort right away

        #---------------------------------------------------------------
        # Verify Account lineups
        # can skip this step with [VerifyLineups] or cmd line -NoVerify
        # queue command if needed
        
        if DoInit and ReadyState <  InitState.SD_LINEUPS :
            
            logging.info("THRD:DoTimerBuild:INIT: Verifying SD Lineup (BuildAddRemove)")
            newLineups = []
            for L in PigSettings['MyLineups'] :
                newLineups.append( L['LineupName'] )
            BuilderQueue.put( {'Command':BuildAddRemove, 'Params':newLineups} )
        
        #---------------------------------------------------------------
        # Build the 'Services' list if needed
        #
        if DoInit and ReadyState <  InitState.SERVICE_RDY :
            logging.info("THRD:DoTimerBuild:INIT: Queuing BuildServices")
            BuilderQueue.put( {'Command':BuildServices} )
            
        #---------------------------------------------------------------
        # Schedule is always queued when build timer fires so build can be recheduled if needed.
        # Also Build Thread needs at least 1 command to run to release SDResourceLock
        # 1.40 Need to build thru services, but not schedule for delayed schedule build
        # in case DVR does a MGR and requests services

        if bDoScheduleBuild :
            logging.info("THRD:DoTimerBuild: Queuing BuildSchedule.")
            BuilderQueue.put( {'Command':BuildSchedule, 'Params':updateMsg } )
        else :
            logging.info("THRD:DoTimerBuild: Skipping Build, Queing Release.")
            BuilderQueue.put( {'Command':BuildRelease } )  # put in at least one command

    return 

def SetWDogBuildTimer(delaySeconds) :

    global WDogBuildTimer, AutoBuildTimeStr
    global gExpectedBuildTimeUTC
    
    # return the local clock time when the timer will fire
    at  = datetime.utcnow()
    
    # If User has specified a fixed build time (fbt), Calc secs to build at that time tomorrow
    
    if PigSettings["StaticBuildTime"] : # in Local time like "07:30" or "15:40"
        #User Hour:Min local time eg 18:35 or 07:00 
        UserBuildTime = parser.parse( PigSettings["StaticBuildTime"] )
        junk, tzmins = LocalTimeStr()
        lt  = at + timedelta(minutes=tzmins) # adjust at to get current local time
        fbt = lt + timedelta(seconds=delaySeconds)  # Advance to next day
        # but reset HH:MM to a user requested build time so no slipage occurs
        fbt  = fbt.replace(hour=UserBuildTime.hour, minute=UserBuildTime.minute, second=0)
        dif  = fbt - lt   # a timedelta object calc exact secons needed from now
        delaySeconds = dif.seconds  # set delay to keep time fixed
        
    # When DoTimerBuild() ends, the DvrBuildTimer thread is dead.
    # We expect a dvr to come along within the next hour & reset this timer, but if it does not,
    # If the dvrs are offline, eg the network is down, the sched could get really stale.
    # The Build timer was set for 23 hrs after the a dvr made a schedule request, we now expect
    # that dvr to make a request within an hour.  We will set a 24 hour timer 'watchdog' timer
    # so that if dvr doesnt ask in the expected time, we will do a build anyway.
    # The WDogTimer will get reset right here each time it fires.

    WDogBuildTimer = threading.Timer(delaySeconds, DoWDogBuildTimer)
    WDogBuildTimer.name = 'Auto Build Timer'
    WDogBuildTimer.daemon = True   # thread is killed when program exits
    WDogBuildTimer.start()
    
    at += timedelta(seconds=delaySeconds)
    AutoBuildTimeStr,tz = LocalTimeStr(at, showSeconds=True)

    at += timedelta(seconds=10)
    gExpectedBuildTimeUTC = at

    logging.info("THRD:SetWDogBuildTimer: Auto Build time scheduled for %s.", AutoBuildTimeStr)

    return AutoBuildTimeStr


def SetDvrBuildTimer(delaysecs) : # returns a string with the Localized Build Time
    global DvrBuildTimer
    global WDogBuildTimer
    global NextBuildTimeStr
    global gExpectedBuildTimeUTC

    if delaysecs == 0 :
        return "DelaySecs=0.  DVR Build Timer not set!"

    # If the auto build timer is set, cancel it since we are scheduling a build
    try :
        if WDogBuildTimer.is_alive() :
            WDogBuildTimer.cancel()
    except :
        logging.warning("THRD:SetDvrBuildTimer: Exception canceling WDogBuildTimer.")

    try :  # if DvrBuildTimer is currently active, kill it and restart with a new time
        if DvrBuildTimer.is_alive() :
            DvrBuildTimer.cancel()
    except :
        logging.warning("THRD:SetDvrBuildTimer: Exception canceling DvrBuildTimer.")

    # have to create a new Timer thread each time, cant reuse or reset thread timers
    DvrBuildTimer = threading.Timer(delaysecs, DoTimerBuild, args=(True,) )
    DvrBuildTimer.name='DVR Build Timer'
    DvrBuildTimer.daemon=True             # thread is automatically killed when program exits
    DvrBuildTimer.start()
    
    bt  = datetime.utcnow()                # return the local clock time when the timer will fire
    bt += timedelta(seconds=delaysecs)
    NextBuildTimeStr,tz = LocalTimeStr(bt, showSeconds=True)

    logging.info("THRD:SetDvrBuildTimer: Build time scheduled for %s.", NextBuildTimeStr)

    # This is a for Sleep/Wake monitoring
    bt += timedelta(seconds=10) # Monitor Real Time 10s passed expected build time
    gExpectedBuildTimeUTC = bt
    
    return NextBuildTimeStr

# Builder Thread Commands will generally use global variables, but some pass in parameters and return results

def Builder() :
    global ReadyState
    global BuilderQueue
    global BuildAbortEvent
    global SDResourceLock
    global AutoSaveCounter
    global RestartServerRequest
    global httpd

    Terminate = False
    BldStat = 0
    CmdToken = None
    
    while not Terminate :  # Builder continues until terminated (nothing terminates it right now)
        
        # Check if someone is requesting a Builder abort (UI save settings can do this)
        if BuildAbortEvent.is_set() :
            logging.info("THRD:BUILDER: Abort Request. Flushing Command Queue.")
            BuildAbortEvent.clear()
            while not BuilderQueue.empty() :
                BuilderQueue.get()
                BuilderQueue.task_done()
        
        # Builder keeps the SDResourceLock until all commands have been processed
        # If Queue Len is 0 and Builder has the lock, then release the SD Resource Lock
        
        if BuilderQueue.empty() and SDResourceLock.locked() :
            logging.info("THRD:BUILDER: Queue is Empty, releasing SDResourceLock\n")
            SDResourceLock.release()

        # Before placing a command in the Queue, the initiator requested the SDResourceLock
        # The caller could have submitted multiple build commands
        # Get a build command from Queue.  this Thread will wait/block here for a command
        
        dBuildCmd = BuilderQueue.get()  # dBuildCmd is Dictionary with 1 or more keys
        buildCmd = dBuildCmd['Command']
        logging.info("THRD:BUILDER: Command is %s",buildCmd)

        #---------------------------------------------------------------
        # First handle non-SD commands that dont need a Token
        
        if buildCmd == BuildTerminate :
            
            logging.info("THRD:BUILDER: Builder Thread Terminating...")
            Terminate = True
            
        elif buildCmd == BuildDelay:
            secs = dBuildCmd['Params']
            logging.info("THRD:BUILDER: Sleeping for %ss",secs)
            time.sleep(secs)

        elif buildCmd == BuildRelease: # A 'Null' command to ensure Lock Relase
            #logging.info("THRD:BUILDER: Sleeping for %ss",secs)
            pass
            
        elif buildCmd == BuildInit:
            #secs = dBuildCmd['Params']
            DoTimerBuild(False) # start a build w/o building schedule
        
        elif httpd and buildCmd == BuildWebserver : # test httpd to ensure webserver exists
            
            logging.info("THRD:BUILDER: Restarting Webserver...")
            
            # restart will happen concurrent with Build Schedule
            
            # signal main loop we want to restart the WebServer
            # (vs ending the PiGS program)
            RestartServerRequest = True
            
            # https://github.com/python/cpython/blob/3.10/Lib/socketserver.py
            # shutdown() seems to do a hard wait locking up PiGS
            # httpd._BaseServer__is_shut_down.set()
            # httpd.shutdown()        # This acts like a Ctrl-c to break main loop
            # so instead set the flags needed to shutdown the HTTP server manually
            #https://www.netjstech.com/2019/04/name-mangling-in-python.html
            httpd._BaseServer__shutdown_request = True

            # Clear event flag to sync restart with this builder - MAIN will set flag when WS restarts.
            WebserverResetEvent.clear()
            
            #pid = os.getpid()
            #signal.pidfd_send_signal(pid, signal.SIGINT , siginfo=None, flags=0) #python 3.9
            #signal.raise_signal( signal.SIGINT ) # send a ctrl-C, python 3.8
            #os.kill(pid, signal.CTRL_C_EVENT)  # 3.2 windows only?
            
            os.kill(os.getpid(), signal.SIGINT)  # send a ctrl-C

            # wait for up to 10 seconds for Main to restart webserver...
            # 
            logging.info("THRD:BUILDER: Waitingfor Webserver to Restart" )
            WebServerReady = WebserverResetEvent.wait(10) 
            
                   
        elif buildCmd == BuildStartDNS:
            
            logging.info("THRD:BUILDER: Starting DNS...")
            
            # ensure we have a PiGS_IP in config to display on UI and for DNS
            if not PigSettings['PiGS_IP'] :  
                logging.info("THRD:BUILDER:DNS: PiGS Does not have an IP!  Checking...", IPaddr )
                IPaddr = WaitForNetwork(1)
                logging.info("THRD:BUILDER:DNS: PiGS Discovered IP = %s", IPaddr )

                if not IPaddr :  # PiGS did not detect an IP address - no Internet Connection?
                    logging.warning("THRD:BUILDER:DNS: Computer does not appear to have an Ethernet connection, cant start DNS" )
                    
                PigSettings['PiGS_IP'] = IPaddr  # save PiGS_IP in config to display on UI

            # Start DNS if we can
            if PigSettings['PiGS_IP'] :  # ensure we have a PiGS_IP to start DNS
                PigsDnsStart(PigSettings['Server_IP'], PigSettings['PiGS_IP'], PigSettings['DnsHosts'])
            else :
                logging.warning("THRD:BUILDER:DNS: Computer does not appear to have an Ethernet connection, cant start DNS" )
                
        else : # its an SD command that needs a token
            
            un = PigSettings['sdusername']     # default: use username and pwd from pigs.json
            pw = PigSettings['sdsha1pwd']
        
            # special case for initial login, un/pw not in PigSettings yet, use params in Command
            if buildCmd == BuildLogin :
                un = dBuildCmd['Params']['un']
                pw = dBuildCmd['Params']['pw']

            # Get a Token For this SD Command
            
            code, tokencode, CmdToken = SDGetToken( un, pw, PigSettings.get('debugToken'))

            # BuildLogin just does a token request.  If success, that proves login
                
            if buildCmd == BuildLogin :  # special case for initial login, return token status
                
                # send the result back as a small dict object
                dBuildCmd['Result'] = {'code':code+tokencode, 'errj':CmdToken}
                dBuildCmd['Event'].set()  #signal waiting process results are ready
            
            # GTR IF  ---------
            if (code == 200 and tokencode == 0) :  # GTR Token get successful -
                
                if buildCmd == BuildLogin :  # nothing to process
                    pass
                
                elif buildCmd == BuildAddRemove :
                    
                    if gVerifyLineups :  # can over-ride this step 
                        try :
                            lineups = dBuildCmd['Params']
                            BldStat = SDAddDeleteLineups(lineups)
                        except :
                            # failed, build timer will be reset below
                            logging.warning("THRD:BUILDER: BuildAddRemove Lineup Exception occurred! ", exc_info=True)
                        else : 
                            if BldStat != 200 :
                                logging.warning("THRD:BUILDER: BuildAddRemove failure (%s)! )", BldStat)
                            else :
                                ReadyState = InitState.SD_LINEUPS
                    else :
                        ReadyState = InitState.SD_LINEUPS

                elif buildCmd == BuildPreview :
                    try :
                        code, lujson, luname = SDGetLineupPreview( dBuildCmd['Params'] )  #
                        # send the result back as a small dict object
                        dBuildCmd['Result'] = {'code':code, 'lujson':lujson,'luname':luname}
                        logging.info("THRD:BUILDER:Preview: Setting callers Event flag.")
                        dBuildCmd['Event'].set()  #signal results are ready
                    except :
                        #e = sys.exc_info()
                        logging.warning("THRD:BUILDER: Build Preview Exception occurred! ", exc_info=True)

                elif buildCmd == BuildServices :
                    
                    if ReadyState >= InitState.SD_LINEUPS :  # then we can build service list

                        try :
                            svcList = BuildServicesListNEW(PigSettings, PiGS_SERVICEID, CmdToken)

                            #if svcList :
                            #    logging.info('THRD:BUILDER:SERVICES: Saving Service List to Disk...')
                            #    utSaveObjectToDisk(svcList, SERVICE_FILE)

                            # for debug, save the json (large) message sent to dvr+
                            if PigSettings.get('SaveJson') :
                                logging.info('THRD:BUILDER: Saving ServiceDetailsMsg to '"'DvrLineup.json'"' ')
                                sdjson = GetServiceDetails(PigSettings, PiGS_SERVICEID, PigSettings['TimeZones'])
                                file = open('DvrLineup.json', "w")
                                file.write(sdjson)
                                file.close()
                                os.chmod('DvrLineup.json', mode=0o777)

                        except :
                            logging.warning("THRD:BUILDER: Build Service Lineup Exception occurred!", exc_info=True)
                        else : 
                            ReadyState = InitState.SERVICE_RDY
                            
                            if PigSettings["UpdateIconCache"] : # if we did an icon cache update, clear the flag
                                logging.info('THRD:BUILDER: Clearing [UpdateIconCache] flag and saving settings...' )

                                PigSettings["UpdateIconCache"] = False
                                # This saves 5 backps as <name>N.json, where N = 1..5
                                SaveSettingsFile(PigSettings)                            
                    else :
                        logging.warning("THRD:BUILDER: Can't Build Service Lineup! ReadyState(%s)", ReadyState)
                            
        
                elif buildCmd == BuildSchedule :
                    
                    # if services was not built, Schedule cant be built
                    
                    if ReadyState >= InitState.SERVICE_RDY :
                        
                        # Disable DNS during schedule builds
                        # Do we really need to stop the DNS if we have a schedule?
                        
                        if PigsDnsAlive() :
                            logging.info("THRD:BUILDER: Suspending PiGS DNS during Schedule Build...")
                            dnsThread = PigsDnsStop(message='THRD:BUILDER:PiGS DNS disabled during Schedule Build...')
                            dnsThread.join(2.0)  # wait up to 2s for DNS to Stop (should stop in less than 1 sec)

                        try :
                            # Set ReadyState state to SCHEDL_NOTRDY 
                            ReadyState = InitState.SCHEDL_NOTRDY  
                            BldStat = BuildLinearSchedule(PigSettings, PiGS_SERVICEID, BuildAbortEvent, dBuildCmd['Params'])
                            # return 0 means schedule was built 
                            # != 0 not built, but yesterdays could still be there
                        except : 
                            logging.warning("THRD:BUILDER: Build Schedule Exception occurred! Schedule Not Updated.", exc_info=True)
                            BldStat = 100  # unexpected exception
                            #LSClearSchedule() 
                        else :
                            if BldStat :
                                logging.warning("THRD:BUILDER: Build did not Complete. Error(%s)! )", BldStat)
                                
                            if LSGetProgramSchedule() :  # if there is a schedule, even from yesterday, something is ready
                                ReadyState = InitState.SCHEDL_RDY
                    else :
                        logging.warning("THRD:BUILDER: Can't Build Schedule! ReadyState(%s)", ReadyState)
                    
            # GTR IF ELSE ---------
            else :  # GTR Token Request failed
                # this might happen if user has not set their SD UN/PW
                logging.warning("THRD:BUILDER:cmd(%s) GetToken Error(%d) %s", buildCmd, code+tokencode, CmdToken)
                BldStat = 200 # no token

            # If PiGS was not able to build the Schedule, schedule a build 20 minutes from now
            # dns resolution failure or SD timeouts may cause build exceptions in SD call requests
                            
            if buildCmd == BuildSchedule :
                
                #if ReadyState == InitState.SCHEDL_RDY :

                # do I have to test if LSGetProgramSchedule()
                
                if LSGetProgramSchedule() and BldStat == 0 :
                    
                    # if build success, schedule a tentative build 24 hours from now
                    SetWDogBuildTimer(24*60*60)
                    
                    # V1.24 AutoSave Config file for changing DNS states
                    if PigSettings["AutoSaveConfig"] : # if a value other than 0 is set....
                        AutoSaveCounter += 1
                        if AutoSaveCounter >= PigSettings["AutoSaveConfig"] :
                            logging.info('THRD:BUILDER: Auto-Saving config settings...' )
                            AutoSaveCounter = 0
                            # This saves 5 backps as <name>N.json, where N = 1..5
                            SaveSettingsFile(PigSettings)                            
                    
                else :  # Build not completed.  Try again in 20 minutes
                    # 20 mins from now (will give 2-3 build chances before dvr needs)
                    # Once the Timer fires and the Build command is Queued, the DvrBuildTimer thread will be dead
                    bts = SetDvrBuildTimer(20*60)
                    logging.info("THRD:BUILDER: BUILD FAILED. Retrying at %s", bts)
                    
                    # Send Email Alert
                    alertmsg = "Build Did Not Complete(%s)!  Retrying at %s.\r\n ." % ( BldStat, bts)
                    SendEmailAlert(alertmsg)
                        
                # Start DNS after build attempt so DVR's dont make requests before Sched is ready
                # ToDo xxx should this be outdented? so we are sure DNS is enabled???
                # If we have a schedule, even yesterdays, we can enable DNS
                # This should also enable DNS for a disk loaded schedule if first build fails
                if LSGetProgramSchedule() and not BuildAbortEvent.is_set() :
                    if PigSettings['PigsDoesDNS'] and not PigsDnsAlive() :
                        BuilderQueue.put( {'Command':BuildStartDNS} )
                    
        logging.info("THRD:BUILDER: ReadyState(%s)", ReadyState)
        BuilderQueue.task_done()
        
        # take a 200mS nap. Other code sometimes submits multiple SD commands.
        # ensure those threads had time to submit all the commands it wanted to
        
        time.sleep(0.500)
        
    logging.info("THRD:BUILDER: Builder Thread Terminated!")
    
    return


#---------------------------------------------------------------------------------------
# References
# https://blog.anvileight.com/posts/simple-python-http-server/
# This reference also discusses SSL support, althought DVR+ does not do SSL
# Subclass because I need to serve some files (index.htm, log file, images) but also
# need to handle the GET and POST requests BEFORE passing on the Simple handler

#Server_error_responses https://developer.mozilla.org/en-US/docs/Web/HTTP/Status
ServerErrorCode = 503  #
ClientBadRequest = 400

gWebServerRestartTime = "Not Started."

class PiGS_HTTPRequestHandler( SimpleHTTPRequestHandler ):

    global bLogHttp
    global PigSettings
    global PigsStartTime
    global gWebServerRestartTime
    global RestartServerRequest


    # Scheudle Builder communication
    global BuilderQueue
    global BuildAbortEvent
    global SDResourceLock
    global BuilderThread
    global ReadyState

    # Scheudle Builder Commands
    global BuildDelay    
    global BuildServices  
    global BuildSchedule 
    global BuildTerminate
 
    # protocol_version defaults to HTTP/1.0
    # One big difference is that in HTTP/1.1 keep-alive is assumed (connection re-use)
    # This seems to have big issues with this code, esp the Pi-Zero. got crashes and lockups
    
#    protocol_version = 'HTTP/1.1'  # set to version 1.21 to support chunking
    # Over-ride to send 'Connection : Close' when serving a file to disable keep-alive
    _close_connection = False

    #override/extend StreamRequestHandler/BaseRequestHandler of socketServer
    def XXhandle(self):  #remove XX to use
        print('httpServer.handle() doing one request')
        SimpleHTTPRequestHandler.handle(self)
        print('httpServer.handle() done with request')


    #-------------------------------------------------------------------
    # over-ride log_message to silence the logging messages sent from
    #      self.send_response()  log_message(format, ...) log_error(format, ...)
    # https://docs.python.org/2/library/basehttpserver.html
    # https://stackoverflow.com/questions/3389305/how-to-silent-quiet-httpserver-and-basichttprequesthandlers-stderr-output

    def log_message(self, format, *args):
        if bLogHttp :                    # a key allows me to block or show these messages
            logging.info(format, *args)  # comment this out to completely silence http:lib
            #logging.warning(format, *args)
            pass
        return

    #-------------------------------------------------------------------
    # Over-ride end_headers() to insert/add 'Connection : Close' when serving a file to disable http/1.1 keep-alive
    # we can flag the Simple file server to close the connection after serving a file
    # internally when a file is sent, we will send the Close header right before the flush
    # see source https://github.com/python/cpython/blob/3.7/Lib/http/server.py
    # the persistent connections were causing all sorts of problems with http.server
    
    def end_headers(self):
        if self._close_connection :
            self.send_header("Connection", "Close")
            #logging.info("MAIN:GET: Adding (Connection : Close) header to response")
            self._close_connection = False   # just do it for this one request
            
        SimpleHTTPRequestHandler.end_headers(self)  # do normal end and flush


    #https://realpython.com/intro-to-python-threading/
    #def __init__(self) :
    #    SimpleHTTPRequestHandler.__init__(self)
    #    super(PiGS_HTTPRequestHandler, self).do_GET()
        
    # self.client_address .server .command .path .request_version .headers)
    
    # https://www.geeksforgeeks.org/http-headers-content-type/
    def SendCommonHeaders(self):
        # HTTP/1.1 200 OK\r\n
        # Headers sent by CM servers to DVR+, We will emulate
        #SimpleHTTP sends these two automatically
        #self.send_header("Date:", "Sat, 02 May 2020 16:04:00 GMT") #Sat, 02 May 2020 16:04:00 GMT\r\n
        #self.send_header("Server:", "nginx/1.10.3 (Ubuntu)")
        self.send_header("Cache-Control", "no-cache,no-store,must-revalidate")
        self.send_header("Pragma", "no-cache")
        self.send_header("Expires", "0")
        self.send_header("Vary", "Accept-Encoding")
#        self.send_header("Connection", "keep-alive")
#        self.send_header("Content-Type", "application/json")
        #These will be set by individual commands
        #self.send_header("Content-Length:", "359")  # size of payload, gzipped data
        #self.send_header("Content-Encoding:", "gzip")

    def do_GET(self):   # I intercept GET's before passing up to parent 'fileserver'

        global PigSettings    # uses the global dictionary
        global DvrBuildTimer
        global NextBuildTimeStr
        global AutoBuildTimeStr
        global RestartServerRequest
                                # https://docs.python.org/3.7/library/urllib.parse.html
        o=urlparse(self.path)   # break url into scheme,netloc,path,params, query,fraqment
        qs=parse_qs(o.query)    # parse the query parameters into a dictionary
        
        #logging.debug("GET: Path: %s\nHeaders:%s", str(self.path), str(self.headers))
        #logging.debug("GET: Path: %s", str(self.path))#, str(self.headers))

        # Handle DVR+ of Get service code from Postal Code
        # The DVR+ zip is going to be ignored - we will return 99999 as the Service Code
        
        # If we wanted to send commands to the server from DVR+, we could 
        # use the zip codes as 'commands'.  eg 5000 could be force SD refresh.
        # For example, set the zip to 50000, then do a guide refresh.
        # The 50000 would be interpretted here and start the download
        # Could require one of these to set the daily download time.
        # after the command, the zip should be set to 99999 on dvr+
        # if user did not reset zip to 99999 the cmd would execute each time dvr did a zip request.
        # We'd need to have a Latch flag (did it), and a reset (another zipcode) for this to work
        
        # Get time of request, convert to local time - easier to match up with what we see on dvr stbhealth page
        doGet_time = datetime.utcnow()  # tor = time of request
        lts, tz = LocalTimeStr(doGet_time, showSeconds=True)
        
        # Development hack to turn DNS on/off w/o doing a full build.  May not play well with the UI
        if ("toggledns" in o.path):
            
            if not PigsDnsAlive() :  # If DNS not running, Start it

                PigSettings['PigsDoesDNS'] = True
                logging.info("MAIN:toggledns: Starting DNS...")
                PigsDnsStart(PigSettings['Server_IP'], PigSettings['PiGS_IP'], PigSettings['DnsHosts'])
                
            else :   # otherwise Stop DNS
                PigSettings['PigsDoesDNS'] = False
                logging.info("MAIN:toggledns: Stopping DNS - waiting for thread to end...")
                dnsThread = PigsDnsStop()
                dnsThread.join(2.0)  # wait up to 2 seconds for DNS to Stop (should stop in less than 1 sec)
                logging.info("MAIN:UIsettings: DNS Stopped... Alive=%s", dnsThread.is_alive())
                
            self.send_response(200)
            self.end_headers()
        
        elif ("postalcode" in o.path): # DVR+ asking for GET of service ID
            
            logging.info("MAIN:**** Postalcode request from %s at %s.", self.client_address, lts )
            
            # Postalcode only happens on a Manual Guide refresh, or when user does some things on the DVR
            # We will cancel the scheduled build time so that the build time syncs with the next dvr???
            
            self.send_response(200)
            self.SendCommonHeaders()
            self.send_header("Content-Type", "application/json")
 
            # We could parse out the postalcode from the URL sent by DRV+ but we are not using it
            # When DVR+ sends a zipcode, we always return a service code of 99999
            # Send the current TimeZone info for incorporation to DVR+
            # Get a json response object to send to DVR+
            
            srjson = GetServiceIdFromZip("", PiGS_SERVICEID, PigSettings['TimeZones'])
            
             # size of payload
            self.send_header("Content-Length", str(len(srjson))) 
            self.send_header("Connection","Close")
            self.end_headers()                      
            self.wfile.write( srjson.encode() )
           
        elif ("servicedetails" in o.path): #DVR+ asking for GET of service ID
            
            # this message can't happen without DVR+ first calling postalcode, so we dont
            # have to check for the build state.
            # Todo, well maybe it could from another DVR????

            logging.info("MAIN:**** Service Details request from %s at %s.", self.client_address, lts )

            # If PiGS Schedule not ready when DVR contacts...
            # if network was not ready at boot, the service list may not have been built
            # Could for SERVICE_RDY, but just want the whole sched ready 

            # if the list does not exist, request a build, send http error, exit
            # if ReadyState != InitState.SERVICE_RDY :
            if not GetServiceList() :
                
                if SDResourceLock.acquire(False) :  # if we got the lock, start the build                
                    logging.warning("MAIN:SERVICEDETAILS: No service list, Initiating build process...")
                    
                    # We have the Lock, we can make SD calls, or Queue commands
                    SDResourceLock.release()  # Build init process will get it's own lock
                    SetDvrBuildTimer(1.0)
                else :
                    # seems like this could never happen - cant be building anything if there is no service list
                    logging.warning("MAIN:SERVICEDETAILS: PiGS is busy rebuilding schedule! (%S)", ServerErrorCode)
                    
                self.send_error(ServerErrorCode, message="PiGS is busy building the Schedule")
                return

            # received headers Accept: */*  Accept-Encoding: gzip,deflate
            self.send_response(200)
            self.SendCommonHeaders()
            
            # Send the Channel lineup list of users Favorite Stations/Channels
            # these have been preconfigured in the UI and are in pigs.json       

            # Get service details and Send to DVR+
            # The ServiceDetails list was built at boot and when user saved settings
            
            # Send DVR+ a service list that was built from SD information
            # This contain the channel numbers, image links, station ID's, etc
            # DVR+ keeps a local copy of this
            
            sdjson = GetServiceDetails(PigSettings, PiGS_SERVICEID, PigSettings['TimeZones'])
            
            # gzip encocded response
            sdgz = gzip.compress( sdjson.encode() )  #to utf-8
            #logging.debug("sdlen(%d)  gzlen(%d)\n", len(sdjson), len(sdgz))

            #gzip encoded response
            self.send_header("Content-Type", "application/json")
            self.send_header("Content-Encoding", "gzip")
            self.send_header("Content-Length", str(len(sdgz))) 
            self.send_header("Connection","Close")
            self.end_headers()
            self.wfile.write( sdgz )

        elif ("linearschedule" in o.path): # DVR+ is asking for a GET of listings

            logging.info("MAIN:LINSCHED **** Program Sched Req from %s at %s", self.client_address, lts )

            # Validate Request Parameters
            # If DVR powers up without an antenna connection (no psip) and it has not done an NTP sync, 
            # it has an unitialized clock time and DVR makes requests like
            #   'startdate': ['1900-00-00T00:00:00Z']  (illegal time)
            #       Exception happened during processing of request from ('192.168.1.201', 41905)
            #       File "/home/pi/pigs/ls.py", line 1178, in GetLinearSchedule
            #       rst     = parser.parse( rsts )
            #       ValueError: month must be in 1..12
            #   'startdate': ['2022-12-28T20:15:55Z']  (strange time)
            #   'startdate': ['1970-01-01T04:00:00Z'] 'startdate': ['1970-01-01T08:00:00Z']  (1970????)
            #if qs['startdate'][0] == '1900-00-00T00:00:00Z' :
            #    logging.warning("MAIN:GET:linearschdule: Bad DateTime... %s", qs['startdate'][0] )
            #    self.send_error(ClientBadRequest, message="Bad StartDate" )
            #    return

            # request start time string, test for wacky year like 1970
            try :
                rst = parser.parse( qs['startdate'][0] )
                if rst.year < 2000 :
                    #logging.warning("MAIN:GET:linearschdule: Bad DateTime... %s (HTTP 400)\n", qs['startdate'][0] )
                    logging.warning("MAIN:GET:linearschdule: Bad DateTime (HTTP 400). Params: %s\n", qs )
                    self.send_error(ClientBadRequest, message="Bad StartDate" )
                    return
            except:
                #logging.warning("MAIN:GET:linearschdule: StartDate Exception... %s (HTTP 400)\n", qs['startdate'][0] )
                logging.warning("MAIN:GET:linearschdule: StartDate Exception (HTTP 400). Params %s\n", qs )
                self.send_error(ClientBadRequest, message="Bad StartDate" )
                return

            # if network was not ready at boot, this list may not have been built
            # minor issue on 1st boot.  If first build fails, a schedule can still exist if loaded
            
            #if ReadyState != InitState.SCHEDL_RDY :    # INIT_COMPLETE :
            if not LSGetProgramSchedule() : # if no schedule built or loaded from disk, then

                if SDResourceLock.acquire(False) :  # if we got the lock, start the build
                    logging.warning("MAIN:LINSCHED: No Schedule built, Initiating scheudule build process...")
                    # We have the Lock, we can make SD calls, or Queue commands
                    SDResourceLock.release()  # Build init process will get it's own lock
                    SetDvrBuildTimer(1.0)
                else :
                    logging.warning("MAIN:LINSCHED: PiGS is busy rebuilding schedule! http(%s)", ServerErrorCode)

                self.send_error(ServerErrorCode, message="PiGS is busy building the Schedule", )
                return
            
            # could happen if another dvr made a request just after the Build Timer fired
            if SDResourceLock.locked() :  # test the SD lockif                
                logging.warning("MAIN:LINSCHED: PiGS is busy rebuilding schedule! http(%s)", ServerErrorCode)
                self.send_error(ServerErrorCode, message="PiGS is busy building the Schedule" )
                return

            # In normal daily updates DVR+ jumps right in with a schedule request since
            #  it already knows its service details

            # This is how the schedule is pre-built on the fly...
            #   Schedule the next build 23 hrs from now, just before this dvr needs it tomorrow
            # The other dvrs will just use the existing schedule b/c its already been built
            # Use the first request from a dvr to schedule the next build time
            
            # If a StaticBuildTime is specified, dont schedule the dynamic 'Prebuild'
            # if this is a DVR's first request of the day, it sets the inprogress flag true
            
            if qs["inprogress"][0]=='true' and not PigSettings["StaticBuildTime"] :   
                if not DvrBuildTimer.is_alive() :  # and the Build timer is not already set (by another DVR)
                    bts = SetDvrBuildTimer(PigSettings["RebuildSecs"]) # then set it for 23 hrs (in seconds) from now
                    #logging.info("MAIN:LINSCHED: Next Build time is scheduled for %s", bts)
                    # Once the Timer fires and the Build command is Queued, the DvrBuildTimer thread will be dead
                    
            
            if not PigSettings['ChunksEnabled'] :
                
                self.send_response(200)
                self.SendCommonHeaders()
                
                # Send doGet_time string for barker channel
                tor, tzx = LocalTimeStr(doGet_time, showSeconds=False, showDate=True, showTimeZone=False, WeekDay=True)
                    
                qs['TOR'] = tor     # add Time of Request to request params for Barker Channel
                qs['BarkerChan'] = PigSettings['BarkerChan']
                
                # Send Linear Schedule Request to DVR+  qs = has the request params sent by DVR+ (plus my Barker info)
                lsjson = GetLinearSchedule(qs, self.client_address[0])
                
                # time.sleep(35)  # testing broken pipe error.  DVR closes connection after 30s

                # This is a very large response - typically 4hours x many channels of data, so compress it
                self.send_header("Content-Type", "application/json")
                self.send_header("Content-Encoding", "gzip")
                lsgz = gzip.compress( lsjson.encode() )  #to utf-8
                #logging.debug("MAIN:LS: lslen(%d)  gzlen(%d)\n", len(lsjson), len(lsgz) )

                self.send_header("Connection","Close")
                self.send_header("Content-Length", str( len(lsgz) )) 
                self.end_headers()
                
                d = datetime.utcnow() - doGet_time
                if d.seconds > 28 :  # DVR closes connection after 30s
                    logging.warning("MAIN:LINSCHED ***** Slow Response getting Request! (%s secs) ***** \n", d.seconds)
                
                try :
                    self.wfile.write( lsgz )
                except BaseException as e:  # need BaseException to catch broken pipe error
                    d = datetime.utcnow() - doGet_time
                    logging.warning("MAIN:LINSCHED Socket write Exception (%s secs). %s\n", d.seconds, e)
                    
            else :  # DONT USE!!!  do Chunked encoding so DVR is not left waiting for a long time
                
                # MUST have protocol_version = 'HTTP/1.1'  at class level for chunked to work
                
                self.send_response(200)
                self.SendCommonHeaders()
                self.send_header("Content-Type", "application/json")
                #self.send_header("Connection", "keep-alive")
                self.send_header("transfer-encoding", "chunked")
                self.send_header("Content-Encoding", "gzip")
                self.end_headers()

#                   Future, do chunking in dedicated function
#                lsjson = GetChunkedLinearSchedule(PigSettings['TimeZones'], qs, PiGS_SERVICEID, \
#                                 self.client_address[0], self.wfile )

                first = True
                ChunkCnt = 1
                Done = False

                #compressobj(level=-1, method=DEFLATED, wbits=15, memLevel=8, strategy=Z_DEFAULT_STRATEGY[, zdict])
                #https://docs.python.org/3.5/library/zlib.html#module-zlib
                #The wbits parameter controls the size of the history buffer (or “window size”),
                #  and what header and trailer format is expected.
                #+25 to +31 = 16 + (9 to 15): Uses the low 4 bits of the value as the window size
                #   logarithm, while including a basic gzip header and trailing checksum in the output.
                
                co = zlib.compressobj(zlib.Z_DEFAULT_COMPRESSION, zlib.DEFLATED, 16+11)
                
                totaldur = int(qs["duration"][0])      # Total duration requested in minutes
                offset = 0
                chunkDur = 30
                if totaldur < chunkDur : chunkDur = totaldur

                #time.sleep(28)  # testing broken pipe error

                while not Done :
                
                    # todo modify qs
                    lsjson = GetChunkedLinearSchedule(PigSettings, qs, offset, chunkDur, PiGS_SERVICEID, \
                                 self.client_address[0], FirstChunk=first )
                    
                    offset += chunkDur
                    if offset > totaldur : offset = totaldur
                    
                    # done if we processed all teh minutes
                    Done = (totaldur == offset)

                    # or if No data was sent
                    if len( lsr["LinearScheduleResult"]["Schedule"]["Airings"] ) == 0 :
                        Done = True

                    # Need to modify the json packet for chunking
                    # first packet search for ']' at end to strip off ] } } }
                    if first :
                        i = -1
                        while i>-10 and lsjson[i] != ']' :
                            i -= 1
                        lsjson = lsjson[0:i]       # strip off the end of the json list ] } } }
                        
                    # middle packet:  add ',' then strip leading '[' and trailing ']]
                    else : # we get an alist - could have [] or [{...},{...}]
                        lsjson = ',' + lsjson[1:-1]
                    
                    # end packet: strip leading [, add comma?, leave trailing ], add } } }
                    if Done :
                        lsjson += '] } } }'
                    
                    # stream compression with zlib
                    # https://docs.python.org/3.5/library/zlib.html
                    enc  = lsjson.encode()   # to bytes
                    lsgz = co.compress( enc ) 
                    
                    # if last chuck of data, flush the streaming compressor
                    if Done :
                        lsgz = lsgz + co.flush()
                    
                    zippedBytes = len(lsgz)
                    
                    if not Done and 0 == zippedBytes : continue
                    
                    logging.info("MAIN:Chunk: %s, lslen(%d) gzlen(%d)\n", ChunkCnt, len(lsjson), zippedBytes )
                    
                    #d = datetime.utcnow() - doGet_time
                    #d.seconds
                    
                    # ascii chunk header - e.g. '1AF\r\n' hex byte count
                    h = '{:X}\r\n'.format(zippedBytes)
                    #h = '{:X}\r\n'.format( len(enc) )
                    
                    # ascii chunk footer & if Done, append '0' length chunk (chunk header, no data, chunk footer)
                    f = '\r\n'  
                    if Done : f += '0\r\n\r\n'  

#                    logging.info("chunk %s %s", self.server.fileno(), self.wfile )
                    ba = h.encode() + lsgz + f.encode()

                    try :
                        self.wfile.write( ba )
                    except Exception as e:
                        d = datetime.utcnow() - doGet_time
                        #d.seconds
                        logging.warning("MAIN:LS %s secs.  Socket write Exception %s\n", d.seconds, e)

                    #time.sleep(0.5)  # testing broken pipe error

                    ChunkCnt += 1
                    first = False
        
            
        elif ("search" in o.path): # user doing a manual schedule search

            logging.info("MAIN:SEARCH: Query=%s", qs)

            self.send_response(200)
            #self.SendCommonHeaders()
            self.send_header("Vary", "Accept-Encoding")
            # with HTTP/1.1 keep-alive is default.  Server tells client it can close connection
            self.send_header("Connection", "Close")

            if not LSGetProgramSchedule() :
                logging.warning("MAIN:SEARCH: Schedule is not built.  Can't search! ")
                htmltxt = "Schedule is not built!  Can't search - See Logs."
            else :
                try :
                    htmltxt = SearchSchedule(LSGetProgramSchedule(), qs )
                    self.send_header("Content-Type", "text/html; charset=UTF-8")
                except:
                    htmltxt = "Search Exception! See Logs."
                    self.send_header("Content-Type", "text/plain; charset=UTF-8")
                    logging.warning("MAIN:SEARCH: Exception occurred! ", exc_info=True)
            
            # gzip encoded
            self.send_header("Content-Encoding", "gzip")
            segz = gzip.compress( htmltxt.encode() )  #to utf-8
            self.send_header("Content-Length", str( len(segz) )) 
            self.end_headers()
            self.wfile.write( segz )

        elif ("showlineup" in o.path): # user dumping lineup to webpage

            logging.info("MAIN:LINEUP: Query=%s", qs)

            self.send_response(200)
            self.send_header("Vary", "Accept-Encoding")
            self.send_header("Content-Type", "text/html; charset=UTF-8")

            try :
                htmltxt = DisplayLineup(GetServiceList(), qs )
            except:
                htmltxt = "Lineup Exception! See Logs."
                self.send_header("Content-Type", "text/plain; charset=UTF-8")
                logging.warning("MAIN:LINEUP: Exception occurred! ", exc_info=True)
            
            # gzip encoded
            self.send_header("Connection","Close")
            self.send_header("Content-Encoding", "gzip")
            segz = gzip.compress( htmltxt.encode() )  #to utf-8
            self.send_header("Content-Length", str( len(segz) )) 
            self.end_headers()
            self.wfile.write( segz )

        elif ("pigstatus" in o.path): #GET of 

            logging.info("MAIN:PIGSTATUS: PiGS Status request from %s at %s.", self.client_address, lts )

            self.send_response(200)
            self.SendCommonHeaders()
            self.send_header("Content-Type", "application/json")

            # add some more PiGS info to the results
            pigstatus = {}

            pigstatus['Generated'] = lts
            
            # PiGS DNS Status 
            pigstatus["PiGS DNS Service"] = PigsDnsStatus()
            
            # Then add the Next Build Time
            msg1 = NextBuildTimeStr if DvrBuildTimer.is_alive() else  "Not Scheduled."
            msg2 = AutoBuildTimeStr if WDogBuildTimer.is_alive() else "Not Scheduled."
            LSBuildStat.SetAlive( BuilderThread.is_alive() )
            LSBuildStat.SetNextBuild(msg1)
            LSBuildStat.SetAutoBuild(msg2)
            pigstatus.update(LSBuildStat._BuildInfo)
                
            pigstatus["DVR Status List"] = '---------------------'
                           
            # Get the status of all DVR's PiGS has updated
            dvrStateD = LsGetDVRStatus( PigSettings["DisplayDVR"]);  #
            pigstatus.update(dvrStateD)  # and merge it into to pigstatus dict

            # Put less frequently used Info at the bottom

            svcStatusDict = GetSVCBuildInfo()
            pigstatus.update(svcStatusDict)  # and merge it into to pigstatus dict
            
            sts, lz = LocalTimeStr( PigsStartTime, True )
            uptime = datetime.utcnow() - PigsStartTime
            pigstatus['PiGS General'] = {}
            pigstatus['PiGS General']["StartTime"] = sts
            pigstatus['PiGS General']["  Up-Time"] = str(uptime)
            pigstatus['PiGS General']["  Version"] = PigSettings['PiGS_Version']
            pigstatus['PiGS General']["Directory"] = os.getcwd()
            pigstatus['PiGS General']["WebServer"] = gWebServerRestartTime

            # convert the status dictionary to text
            dvrjson = json.dumps( pigstatus, indent=2 )

            # Save the Status to a file so users can eMail it to me
            file = open('pigstatus.json', "w")
            file.write(dvrjson)
            file.close()
            os.chmod('pigstatus.json', mode=0o777)
            
            self.send_header("Connection","Close")
            #unencoded response 
            self.send_header("Content-Length", str( len(dvrjson)) ) 
            self.end_headers()
            self.wfile.write( dvrjson.encode() )
        
        elif ("UIsettings" in o.path): #GET of 

            logging.info("MAIN:GET: UIsettings... from %s", self.client_address) #, lts)

            self.send_response(200)
            self.send_header("Content-type", "application/json")

            # export dict to json string, then encode str to bytes object
            jstr = json.dumps( PigSettings )   # Send everything to the UI
            
            self.send_header("Connection","Close")
            self.send_header("Content-Length", str( len(jstr) )) 
            self.end_headers()
            
            self.wfile.write( jstr.encode() )
            return
        
        elif ("lineupPreview" in self.path) :  # UI requesting a SD Lineup Preview based on a Zip

            # Preview needs the SD resource lock b/c it uses the SD token.
            # if a schedule build was in progress this UI call would kill the build

            if not SDResourceLock.acquire(False) :  # if we get the lock, we can do the preview
                logging.warning("MAIN:PREVIEW: Can't Preview.  PiGS is busy rebuilding schedule!")
                self.send_error(ServerErrorCode, message="PiGS is busy building the Schedule. Please Retry.")
                return

            # Build a LineupDef to send to the UI fo display
            # LineupDef = {
            #   "LineupName" : "USA-OTA-53593",   # SD Name for the lineup
            #   "ChannelList": []                 # List of ChannelDefs.
            # }                                   # It will include every channel in the lineup

            # Builder will signal us when Preview data is ready
            PreviewReadyEvent = threading.Event() 
            PreviewReadyEvent.clear()

            # create a build command 
            pcmd = copy.copy(BldCmd)
            pcmd['Command'] = BuildPreview
            pcmd['Params']  = qs['postalcode'][0]
            pcmd['Event']   = PreviewReadyEvent
            
            logging.info("MAIN:GET:lineupPreview: Queuing BuildPreview for %s.  Waiting for Result...", qs['postalcode'][0])
            BuilderQueue.put( pcmd )
            
            # this command could take a long time if the Preview has to be built by the server.
            # it might not even complete on the SD server in time, the user just has to retry.
            # The PiGS webserver is busy during this wait and cant respond to other commands
            
            PreviewReady = PreviewReadyEvent.wait(60) # wait here for up to 60 seconds
 
            if PreviewReady :
                logging.info("MAIN:GET:lineupPreview: Result is ready.")
                # Get the results
                d = pcmd['Result']   # get reference to the result dictionary 
                code   = d['code']
                lujson = d['lujson']
                luname = d['luname']
            else :
                logging.warning("MAIN:GET:lineupPreview: Timeout waiting for Preview Result.")
                code = 503
                lujson = None
                luname = "Timeout"
            
            # Call Lineup Preview command which makes 2 SD calls to get the data.
            # Request the lineup preview. we can get a 200 with an error block with a code in it
            # There is also a differnt type of error response with a non-zero code.
            # we dont know which we will get
                       
            if code == 200 :
                # check for error code in json data, it might be a 2107 Queued not ready error
                # check for an error code, a successful preview response doesnt have an error code
                
                resultcode = 0  # default in case result is a list[]
                
                # SD returns List[] for preview or dict{} for error / status response
                #  cant do a .get() on a list so test for structure type
                
                if type(lujson) is dict :
                    resultcode = lujson.get('code',0)
               
                if resultcode == 0 :
                    
                    # The returned lujson IS the LineupDef ChannelList, but the ChannelDef's
                    #   dont include the favorite key
                    # Create the 'favorite' key and default to True for all channels
                    for c in lujson :
                        c['favorite']=True
                        
                    # create the complete response message to send to the UI
                    msg = copy.copy(LineupDef)    # Start with a copy of my custom response
                    msg['LineupName']  = luname;  # Add the SD lineup name (for display on the UI side)
                    msg['ChannelList'] = lujson;  # append the lineup list of ChannelDefs with Fav key
                    msgj = json.dumps(msg)        # dict to json
                else :
                    logging.info("MAIN:PREVIEW: return code %d", resultcode)                
                    msg = {'code':2107, 'message':"The lineup is being generated, please retry."}
                    msgj = json.dumps(msg)
            else :
                logging.info("MAIN:PREVIEW: Error %d %s %s", code, lujson, luname)                

                
            if lujson == None :
                msg = {'code':code, 'message': luname }
                msgj = json.dumps(msg)

            # forward the SD response code to the UI
            self.send_response(code)
            self.SendCommonHeaders() 
            self.send_header("Content-Type", "application/json")
            self.send_header("Connection","Close")
        
            # gzip encocded response
            lugz = gzip.compress( msgj.encode() )  #to utf-8
            
            self.send_header("Content-Encoding", "gzip")
            self.send_header("Content-Length", str( len(lugz) )) 
            self.end_headers()

            logging.debug("  GET lineupPreview: lulen(%d)  gzlen(%d)\n", len(msgj), len(lugz) )

            self.wfile.write( lugz )
            return
             
        else: # we have not handled this GET.  Let SimpleHTTP handle file requests
            
            logging.info("MAIN:GET: File (%s) from %s %s", o.path, self.client_address, lts)
            # if it sees '/', it looks for and returns index.htm (or index.html)
            # 2 ways to call the Parent... name the parent, or super the child...
                 #super(PiGS_HTTPRequestHandler, self).do_GET()
           
            self._close_connection = True  # Add a 'Connection : Close' header 
            
            SimpleHTTPRequestHandler.do_GET(self)

            logging.debug("MAIN:do_GET Path: %s\nHeaders:\n%sRequestLine:%s\nMAIN:do_GET:Done.", str(self.path), str(self.headers), self.requestline)
 
    def do_POST(self):  # this receives URLs like POST /UIsettings or /sdlogin
        
        global PigSettings

        # time of request, convert to local time - easier to match up with what we see on dvr stbhealth page
        doPost_time = datetime.utcnow()
        lts, tz = LocalTimeStr(doPost_time, showSeconds=True)

        
        # Break url into scheme, netloc, path, params, query,fraqment
        # then parse the query parameters into a dictionary
        o=urlparse(self.path)
        qs=parse_qs(o.query)

        # get and decode the body json, can we assume there is content and a header????
        
        content_length = int(self.headers['Content-Length'])

        #if content_length == 0 :
            
        body = self.rfile.read(content_length)
        
        logging.debug("MAIN:POST: CLEN: %d, Path: %s Headers: %s", content_length, str(self.path), str(self.headers))

        # generic decode for error return/debug, may be re-decoded for specific command
        # bj = json.loads( body.decode() )  #xxx
        
        # Now process the individual Post commands
        
        if ("UIsettings" in self.path) :  # POST settings
           
            respcode = 200
            
            if content_length == 0 :
                respcode = 418 # Tea pot
            else : 
                try :
                    PigsUIreturned = json.loads( body.decode() )  # attempt json decode
                except:
                    respcode = 400  # Bad request
                    
            if respcode != 200 :

                logging.warning("MAIN:POST:sdlogin Bad Request Content Len = %d ", content_length )

            else : # if respcode == 200 :

                # Settings returned from the UI
                # ON Save Settings, lineup or icons may have changed we need to abort any build in progress
                # If a schedule build was in progress, UI Save would kill the build b/c the token would change.
                
                #Todo disable UI button until this command returns???
                
                if not SDResourceLock.acquire(False) :  # if we get the lock, we can do the settings
                    
                    logging.warning("MAIN:POST:UIsettings: PiGS is busy building schedule.  Requesting Abort...")
                    
                    BuildAbortEvent.set()   #signal Linear Schedule Builder to abort
            
                    time.sleep(1)    # give thread time to end
                    
                    waitAttempts = 5 #seconds
                    while waitAttempts and not SDResourceLock.acquire(False) :
                        time.sleep(1)
                        waitAttempts -= 1

                    BuildAbortEvent.clear() # ensure flag is clear after abort
                    
                    if waitAttempts == 0 :
                        logging.warning("MAIN:POST:UIsettings: Timeout on Abort. PiGS could not save Settings!")
                        self.send_error(ServerErrorCode, message="PiGS is busy building the Schedule")
                        #We never got the lock so I dont have to release it, but better to be safe
                        #if SDResourceLock.locked() :
                        #    SDResourceLock.release()
                        
                        #self.send_response(513)
                        #self.send_header("Connection","Close")
                        #self.end_headers()
                        self.send_error(ServerErrorCode, message="PiGS did not respond to Abort.  Please Retry.")
                        
                        return  # xxx what about return code?????
                    
                # We have the Lock, we can make SD calls, or Queue commands
                SDResourceLock.release()  # Build init process in DoTimerBuild will get it's own lock
                
                # If this function returns before Queuing the Service build, must do a SDResourceLock.release()
                
                #--------------------------------------------------------------
                # Get the settings from the UI message block, json to dict conversion
                #PigsUIreturned = json.loads( body.decode() )
                
                logging.info("MAIN:POST:UIsettings (%s bytes) at %s.", content_length, lts )
                
                self.send_response(200)
                self.send_header("Connection","Close")
                self.end_headers()

                #--------------------------------------------------------------
                # I could just replace the whole PigSettings dict with the received UI json.
                #   but maybe there are some keys we want to preserve on the server side
                # I do this with keys in pigsPreserveSettings.  See notes where dict is defined.
                # Copy some keys from current pigs settings to received settings
                # PigsPreserveSettings is just a list of keys only
                # Update the returned keys with current keys in case a few special keys changed while UI opened
                
                for key in PigsPreserveSettings :  # Use local definition structure
                   PigsUIreturned[key] = PigSettings[key]
                
                # If icon type changes or SD theme changes, we will update Cached Icons
                #IconChanged  = ( PigsUIreturned['IconType']    != PigSettings['IconType'] )
                #IconChanged |= ( PigsUIreturned['SdIconTheme'] != PigSettings['SdIconTheme'] )
                #logging.info("MAIN: Icon Change State: %s", IconChanged )

                # Save change flag in config file so Service build can update icons.
                # after service build, should clear this flag and save settings
                
                #PigsUIreturned['UpdateIconCache'] = IconChanged  # service build will update Icon Cache
                

                # Do I need to sanity check the data ???
                # should I check the data before saving?
                # should I make the changes on the Temp settings like do the SD stuff first?
                # Just save what the User configured, if there is a fault at least we saved users work
                # when they restart, refresh UI and click save settings all should be well

                if PigsUIreturned['LoggingLevel'] != PigSettings['LoggingLevel'] :
                    logging.info("MAIN: Logging level changed from %s to %s", PigSettings['LoggingLevel'], PigsUIreturned['LoggingLevel'] )
                    
                
                #--------------------------------------------------------------
                # Recursively copy the UI data into our main config dict
                PigSettings = copy.deepcopy(PigsUIreturned)

                logging.info("MAIN: Saving DVR_States...")
                PigSettings['DVR_States'] = LS_GetDictOfDVRs()

                # This saves 5 backps as <name>N.json, where N = 1..5
                SaveSettingsFile(PigSettings, CONFIG_FILE)

                #--------------------------------------------------------------
                #https://docs.python.org/3/library/logging.html
                #See the heading 'Module Level Functions'
                #Get for a reference to the default root level logger
                logger = logging.getLogger()
                logger.setLevel( PigSettings['LoggingLevel'] )
                
                
                #--------------------------------------------------------------
                # Process PiGS DNS Server changes
                #global IPaddr
                
                # Starting DNS
                if PigSettings['PigsDoesDNS'] and not PigsDnsAlive() :  # do we need to start DNS?
                    if PigSettings['PigsDelayDNS'] :
                        logging.info("MAIN:UIsettings: DNS will be started after Schedule is built...")
                    else :
                        logging.info("MAIN:UIsettings: Starting DNS...")
                        PigsDnsStart(PigSettings['Server_IP'], PigSettings['PiGS_IP'], PigSettings['DnsHosts'])
                
                # Stopping DNS
                if not PigSettings['PigsDoesDNS'] and PigsDnsAlive() :
                    logging.info("MAIN:UIsettings: Stopping DNS - waiting for thread to end...")
                    dnsThread = PigsDnsStop()
                    dnsThread.join(2.0)  # wait up to 2 seconds for DNS to Stop (should stop in less than 1 sec)
                    logging.info("MAIN:UIsettings: DNS Stopped... Alive=%s", dnsThread.is_alive())

                #--------------------------------------------------------------
                # Process Lineup changes on Schedules direct
                # We need to make SD match what user configured by deleting and adding lineups
                # After User Saves settings we do a complete re-initialization
                
                logging.info("MAIN:POST:UIsettings: Clearing ReadyState to Re-initialize...")
                ResetReadyState()
                
                bts = SetDvrBuildTimer(1.5)  # set timer for 1.5 second to allow user to switch window
                
    #            logging.info("MAIN:POST:UIsettings: Build time scheduled for %s", bts)
                # Done processing UI settings

        # User hits 'Verify Login' to create a sha1 password hash, Test login by getting SD token
        # body consists of a small json object (dictionary) containing the username and password
        #   var params = {username: "", password:""};
        
        elif ("sdlogin" in self.path) :  

            respcode = 200
            
            if content_length == 0 :
                respcode = 418 # Tea pot
            else : 
                try :
                    bj = json.loads( body.decode() )  #xxx
                except:
                    respcode = 400  # Bad request
                    
            if respcode != 200 :

                logging.warning("MAIN:POST:sdlogin Bad Request. Content Len = %d ", content_length )
               
            else : # if respcode == 200 : # continue if we have valid content

                sha1pwd = GetSha1Password(bj['password'])
                
                logging.info("MAIN:POST:SDLogin Attempting GetToken")

                if not SDResourceLock.acquire(False) :  # if we get the lock, we can do the sdlogin
                    logging.warning("MAIN:POST:SDLogin Can't login.  PiGS is busy rebuilding schedule!")
                    self.send_error(ServerErrorCode, message="PiGS is busy.  Please Retry.")
                    return  # didnt get the lock, dont have to release anything

                # Builder will signal us when Login data is ready
                LoginReadyEvent = threading.Event() 
                LoginReadyEvent.clear()

                # create a build command
                lgcmd = copy.copy(BldCmd)
                lgcmd['Command'] = BuildLogin
                lgcmd['Params']  = {'un':bj['username'],'pw':sha1pwd }
                lgcmd['Event']   = LoginReadyEvent
                
                logging.info("MAIN:POST:SDLogin: Queuing BuildLogin & waiting for Result...")
                BuilderQueue.put( lgcmd )

                # The PiGS webserver is busy during this wait and cant respond to other commands

                # Increased to 90s in 1.39 because SD does retries for up to a minute
                
                LoginReady = LoginReadyEvent.wait(90)   # wait here for up to 90 seconds
                
                if LoginReady :
                    logging.info("MAIN:POST:SDLogin: BuildLogin is ready.")
                    d = lgcmd['Result']   # get reference to the result dictionary 
                    respcode  = d['code']
                    rsj       = d['errj']
                else :
                    logging.warning("MAIN:POST:SDLogin: Timeout waiting for BuildLogin Result.")
                    respcode = 503
                    rsj  = None
               
                # if it works, the username/sha1 combo is good and we can save in the config file
                if respcode == 200 :
                    # update the config dictionary
                    logging.info("MAIN:POST:SDLogin: SD Login Successful!")
                    PigSettings['sdusername'] = bj['username']
                    PigSettings['sdsha1pwd']  = sha1pwd
                    SaveSettingsFile(PigSettings, CONFIG_FILE)     # Save setting file to disk
                else :
                    logging.warning("MAIN:POST:SDLogin: code(%s) %s", respcode, rsj)
            
            self.send_response(respcode)  # 
            self.send_header("Connection","Close")
            self.end_headers()
            
        else : # its an unrecognized POST command
            #logging.warning("MAIN:POST: Unsupported command! Path: %s Headers: %s body: %s", str(self.path), str(self.headers), body.decode() ) 
            logging.warning("MAIN:POST: Unsupported command! Path: %s body: %s", str(self.path), body.decode() ) 
            #self.send_error(ServerErrorCode, message="PiGS did not respond to Abort.  Please Retry.")
            respcode = 404  # Not Found
            self.send_response(respcode)  
            self.send_header("Connection","Close")
            self.end_headers()

    def _do_unsupported(self):
        logging.warning("MAIN:Unsupported Command: do_ %s", self.command)
        logging.warning("MAIN: Path: %s\nHeaders:%s", str(self.path), str(self.headers))
        self.send_response(400)
        self.send_header("Connection","Close")
        self.end_headers()
        response = BytesIO()
        response.write(b'This is an unsupported request. ')
        
    #https://stackoverflow.com/questions/43156023/what-is-http-host-header
    #https://bugs.chromium.org/p/chromium/issues/detail?id=521009
    #https://superuser.com/questions/1398098/how-can-i-get-google-chrome-to-go-to-an-internal-page-without-performing-an-exte
    #https://docs.microsoft.com/en-us/troubleshoot/browsers/one-word-intranet-sites
    def do_HEAD(self):
        logging.info("MAIN:do_Head. Passing to parent.")
        self._close_connection = True  # Add a 'Connection : Close' header 
        SimpleHTTPRequestHandler.do_HEAD(self)
        logging.debug("MAIN:do_HEAD Path: %s\nHeaders:\n%sRequestLine:%s\nMAIN:do_HEAD:Done.", str(self.path), str(self.headers), self.requestline)
    def do_DELETE(self):
        self._do_unsupported()
    def do_PATCH(self):
        self._do_unsupported()
    def do_OPTIONS(self):
        self._do_unsupported()
    def do_PUT(self):
        self._do_unsupported()

#---------------------------------------------------------------------------------------
# Function to get my IP for display and wait for network to boot
# When starting PiGS at boot, the network is not always ready and
# returns: [Errno 101] Network is unreachable
# We will sleep a few times while waiting for network to boot
# 30 seconds max wait = 30 attempts, 1 sec wait each time
# 
# None is the return value if cant get IP

import socket

def WaitForNetwork(NetworkAttempts = 30):

    for i in range(NetworkAttempts, -1,-1) : # start, stop 1 before, step
    
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        try:
            # doesn't even have to be reachable
            s.connect(('10.255.255.255', 1))
            IP = s.getsockname()[0]
        except Exception as err:
            logging.debug("  get_ip = %s", err )
            IP = None
        finally:
            s.close()
    
        if IP :        # If we get real IP, break out of wait loop
            break
        
        logging.debug("  Waiting for network %d", i )
        time.sleep(1)      # take a nap while the system boots the Network 

    #time.sleep(1)      # take a nap while the system boots the Network 

    return IP



        
##########################################################################
# OK, here we go... initialize and start things going

#====================================================================================
# Process command line args.
# Pass in Port and maybe logging level to use config for logging level
# On Pi/Linux Port 80 is 'protected' so PiGS must be started with the sudo command
#  to allow the app to 'bind' to port 80.  No error checking is done here
#   sudo python3 pigs.py 

#https://docs.python.org/3/howto/argparse.html
import argparse
cmdparser = argparse.ArgumentParser()
#group = cmdparser.add_mutually_exclusive_group(required=False)
cmdparser.add_argument("port",           help="Specify HTTP port other than default port 80", nargs='?', type=int, default=80)
cmdparser.add_argument("-serverIP","-s", help="Set Server_IP. PiGS will listen to ports 53/80 on this IP only")
cmdparser.add_argument("-loglevel","-l", help="Set logging level. L=10 Debug, L=20 Info L=30 Warning", type=int)
cmdparser.add_argument("-force",   "-f", help="Force port to override IDE_Port for debug", action="store_true")
cmdparser.add_argument("-ignore",  "-i", help="Ignore 'No Ethernet Connection' error and continue startup.", action="store_true")
cmdparser.add_argument("-waitnet", "-w", help="Wait 'waitnet' seconds for Network to become available.", type=int)
cmdparser.add_argument("-noload",  "-n", help="Do NOT load Saved Program Schedule or Cache from disk on start", action="store_true")
cmdparser.add_argument("-noverify","-v", help="Skip SD Lineup Verification at Startup", action="store_true")
cmdparser.add_argument("-api",     "-a", help="Set an alternate SD server or API")
#cmdparser.add_argument("-build",   "-b", help="Set first build delay in HOURS, e.g. 18.5", type=float)
cmdparser.add_argument("-delay",   "-d", help="Set first build delay in HOURS, e.g. 18.5", type=float)
cmdparser.add_argument("-holdoff", "-o", help="Holdoff DNS until after first build delay to Sync DVR's", action="store_true")
cmdparser.add_argument("-update",  "-u", help="Add/Update new config keys to pigs.json and exit", action="store_true")
# block dns, sync, synchronize, 

args = cmdparser.parse_args()

#Group Mutually exclusive doesnt work for me.  doing it myself
#usage: pigs.py [-h] [-p P] [-loglevel LOGLEVEL] [port]
#pigs.py: error: one of the arguments port -p is required
#if args.p ('p' in args) and 'port' in args :
#    cmdparser.print_help()
#    cmdparser.print_usage()   # print brief command line format
#    cmdparser.exit(status=0, message='pigs.py: error: Use 1 or none of the arguments port -p, but not both\n')
#    print('port in args')

print("\nPiGS Starting.  No logging until config file loaded...")
print("Command line args:", args)

# Some print()'s are used so something appears when logging to a file

#https://docs.python.org/3/howto/logging.html
#https://docs.python.org/3/library/logging.html#levels
#https://docs.python.org/3/library/logging.html#logging.basicConfig
#https://docs.python.org/3/library/logging.html#module-level-functions
#https://docs.python.org/3/library/logging.html
#See the heading Module Level Functions

#Level      When it’s used
#NOTSET      0
#DEBUG      10 Detailed information, typically of interest only when diagnosing problems.
#INFO       20 Confirmation that things are working as expected.
#WARNING    30 An indication that something unexpected happened, or indicative of some problem in the near future (e.g. ‘disk space low’). The software is still working as expected.
#ERROR      40 Due to a more serious problem, the software has not been able to perform some function.
#CRITICAL   50 A serious error, indicating that the program itself may be unable to continue running.
#The default level is WARNING

# Open CONFIG_FILE JSON file, load settings into a Dict
# Cant do any logging before the config file since we get the log file name from there
#https://stackoverflow.com/questions/9823936/python-how-do-i-know-what-type-of-exception-occurred
#https://docs.python.org/2/library/exceptions.html#exception-hierarchy

bDelayedLogging = False

print("PiGS Working Directory on launch:" , os.getcwd())

# V0.99 Change PiGS working directory to where program started
head, tail = os.path.split( os.path.realpath(__file__) )
os.chdir(head)

#====================================================================================
# Test for icons folders, create if not found.
# the mode params do not seem to work for mkdir so using chmod

if not os.path.exists('myicons'):
    print("Folder 'myicons' does not exist.  Creating...")
    os.mkdir('myicons')  #, mode=0o644, *, dir_fd=None)
    os.chmod('myicons', mode=0o777)
if not os.path.exists('cmicons'):
    print("Folder 'cmicons' does not exist.  Creating...")
    os.mkdir('cmicons')  #, mode=0o644, *, dir_fd=None)
    os.chmod('cmicons', mode=0o777)
if not os.path.exists('sdicons'):
    print("Folder 'sdicons' does not exist.  Creating...")
    os.mkdir('sdicons')  #, mode=0o644, *, dir_fd=None)
    os.chmod('sdicons', mode=0o777)


#====================================================================================
# Load the Config file or create one

from ut import PigsPath

cfgf = PigsPath(CONFIG_FILE)
    
try:
    f = open(cfgf)               # 'pigs.json'
    PigSettings = json.load(f)   # returns JSON object as a dictionary 
    f.close()                    # Closing file

#except json.JSONDecodeError:  
     #JSONDecodeError exception isnt in Python 3.4 and earlier
     # will catch below as generic exception
#    print("Error decoding config file.", CONFIG_FILE)
#    sys.exit(5)

except FileNotFoundError:
    print("Config file %s not found, creating...", cfgf)
    PigSettings = copy.copy(PigsConfigDict)
    with open(cfgf, "w") as outfile: 
        json.dump(PigSettings, outfile, indent=2)
        outfile.close()
    # need to delay logging this b/c logging not enabled yet - log file name in config file
    bDelayedLogging = True
    # We can contine, but this is pretty serious unless it is the very first run
    # I guess its not too serious...  if its gone, you just have to reconfigure.
    
except Exception as ex:
    template = "An exception of type {0} occurred. Arguments:\n{1!r}"
    message = template.format(type(ex).__name__, ex.args)
    print(message)
    print('If you edited pigs.json, possible syntax error (comma or quote?) - use backups pigs1-5.json?')
    sys.exit(2)

# update the version number in the config file in case the app version has changed.
PigSettings['PiGS_Version'] = PIGS_VERSION


#====================================================================================
# Auto-Upgrade pigs.json with missing keys when there are version upgrades

dirty = UpgradeConfigFile()     # returns True if any keys added or removed

if dirty :  # dirty flag added in 1.40
    print('User Settings updated to current version.  Saving...')
    SaveSettingsFile(PigSettings, cfgf)
elif args.update :
    print('User Settings are up-to-date...' )
    

if args.update :
    #SaveSettingsFile(PigSettings, cfgf)
    print('Exit(0) due to Update Exit flag... ' )
    sys.exit(0)

#====================================================================================
# Init Email alerts with Smtp login and destination emails

EM_Initialize(PigSettings['SmtpUser'], PigSettings['SmtpPass'], PigSettings['AlertEmails'], PigSettings['PiGS_IP'])

#====================================================================================
# Now have the configuration file.  http port over-ride?

port = args.port  # default 80 or what user specifies on cmd line

if not args.force and PigSettings["IDE_Port"] > 0 :
    port = PigSettings["IDE_Port"]
    print("IDE_Port Over-ride: %s", port)


#====================================================================================

gVerifyLineups = PigSettings["VerifyLineups"]

if args.noverify :  # if flag is set, skip the lineup Verification
    
    gVerifyLineups = False


#====================================================================================
# Get logging settings from config file, then Enable Logging

logFile  = PigSettings.get('LoggingFile',  PIGS_LOG_FILE)  # if logFile='' it will log to console
logLevel = PigSettings.get('LoggingLevel', logging.INFO)
logf     = PigsPath(logFile)

# Command line over-ride for this sesson only.  UI can change
if args.loglevel :
    logLevel = args.loglevel

# if running from the Dev environment, dont log to file so as not to
#   conflict with the port 80 process if it is running.
# running both at the same time could have side effects with the config file also

if port != 80 :     
    #logf = ""      # if not port 80, remove the filename to log to console only
    # if not port 80 append port to file name so diff pigs instances have their own logs
    logf = PigsPath("pigslog"+str(port)+".txt")
    
# replaced this code with the RotatingFileHandler
#https://stackoverflow.com/questions/24505145/how-to-limit-log-file-size-in-python
#https://gist.github.com/amamaenko/653743
# initialize logging before making calls to any modules that use logging or weird(nothing) thing happen

logger = logging.getLogger()

# add a Rotating File Handler to keep log file size in check 

LogMaxBytes = PigSettings.get("LogFileBytes", MAX_LOGFILE_BYTES)
LogBackups  = PigSettings.get("LogFileCount", LOGFILE_BACKUPS)

if logf :
    try :
        handler = logging.handlers.RotatingFileHandler(logf, maxBytes=LogMaxBytes, backupCount=LogBackups)
        logger.addHandler(handler)
        logger.setLevel( logLevel )
    except PermissionError:
        print('\nError - could not open log file. Insufficient permissions.')
        print("   On a Raspberry Pi or Linux you may need to start as 'sudo python3 pigs.py' (run as root)\n")
        sys.exit(1)
else :
    logger = logging.basicConfig(filename=logf,level=logLevel)

# Also send Info logs to the console
# docs.python.org/2.4/lib/multiple-destinations.html
console = logging.StreamHandler()
console.setLevel(logging.INFO)
#console.setLevel(logLevel) #

#formatter =logging.Formatter('%(name-12s: %levelname)-8s %(message)s')
#console.setformatter(formatter)

#if port != 8000 :     # again, if command line start, log to console also
#    logging.getLogger("").addHandler(console)
logging.getLogger("").addHandler(console)

loggingto = 'file: ' + logf if logf else 'console ONLY <<<<<<<<<<<<<<<<<<<<<<<<<'
print("Logging to ", loggingto)
print("Logging Level is", logLevel)

# set global flag to over-ride annoying HTTP.server log.info prints
bLogHttp = PigSettings.get('LogHttpReq')


# block the "Starting new connection..." requests message
logging.getLogger("requests").setLevel(logging.WARNING)
logging.getLogger("urllib3").setLevel(logging.WARNING)

PigsStartTime = datetime.utcnow()
lst, lz = LocalTimeStr( PigsStartTime, WeekDay=False )
logging.critical("MAIN: PiGS started at %s", lst )
logging.critical("MAIN: Logging Level is %s", logLevel)
logging.critical("MAIN: PiGS Version %s", PigSettings['PiGS_Version'] )
logging.critical("MAIN: Current Working Directory %s" , os.getcwd())

# Had to delay logging this because the logfile did not yet exist when error occurred
if bDelayedLogging :
    logging.warning("MAIN: Config file %s not found, created!", cfgf)

#====================================================================================
#
logging.info("MAIN: Platform Info")
logging.info("  MaxSize(%s), Is64-bit(%s)", hex(sys.maxsize), sys.maxsize > 2**32 )

try :
    import platform
    logging.info("  Achritec: %s", platform.architecture() )
    logging.info("  System  : %s", platform.system() )
    logging.info("  Machine : %s", platform.machine() )
    logging.info("  Platform: %s", platform.platform() )
    logging.info("  Python I: %s", platform.python_implementation() )
    logging.info("  Python V: %s", platform.python_version() )
except :
    logging.warning("  Platform library not available")
    
#Python Version as a Float for comparisons
PythonVer = VersionStringToFloat(platform.python_version())

if PigSettings["ThreadedHTTP"] and PythonVer < 3.7 :
    print("MAIN:WARNING: PiGS 1.24 and up requires a minimum of Python 3.7!" )
    print("MAIN:WARNING: Update Python with 'sudo apt install python3.7'\n" )
    logging.warning("MAIN:WARNING: PiGS 1.24 and up requires a minimum of Python 3.7!" )
    logging.warning("MAIN:WARNING: Update Python with 'sudo apt install python3.7'\n" )
    # Exit pigs, no sense in running pigs if we dont have the required Python Version
    sys.exit(5)

logging.info("MAIN: http.server Info")
try :
    logging.info("    server_version: %s", PiGS_HTTPRequestHandler.server_version )
    logging.info("       sys_version: %s", PiGS_HTTPRequestHandler.sys_version  )
    logging.info("  protocol_version: %s", PiGS_HTTPRequestHandler.protocol_version)
except :
    logging.warning("MAIN: Error getting http.server version info.")


#====================================================================================
# Wait for Network Ready and Print the Pigs IP address.

waitnet = PigSettings['NetworkWait']

if args.waitnet :
    waitnet = args.waitnet

logging.info("MAIN: Checking Network.  Waiting up to %ss for Network ready...", waitnet )
IPaddr = WaitForNetwork(waitnet)
logging.info("MAIN: PiGS Discovered IP = %s", IPaddr )

if not IPaddr :  # PiGS did not detect an IP address - no Internet Connection?
    logging.warning("MAIN: Computer does not appear to have an Ethernet connection." )
    
    if not args.ignore :  # PiGS needs a network.  can override with -ignore
        #cmdparser.print_help()  # prints full help
        cmdparser.error('Fix network connection and retry.')    

PigSettings['PiGS_IP'] = IPaddr  # save PiGS_IP in config to display on UI

#====================================================================================
# manual over-ride of listen IP address for port 80 & port 53
# Must happen after Network detection above

from pigsdns import ValidateIPaddress

hostIP = PigSettings["Server_IP"]    # defualt "0.0.0.0",

if args.serverIP :  # Cmd line override of ip
    hostIP = args.serverIP 
    logging.info("MAIN: Server_IP Command line Override: %s", hostIP )

if not ValidateIPaddress(hostIP) :
    logging.critical("MAIN: *** Server_IP(%s) Validation Error! *** ", hostIP )
    sys.exit(1)

if hostIP != '0.0.0.0' :
    
    PigSettings['PiGS_IP'] = hostIP  # Display on UI
    
    # Print a warning if my discovered IP does not match the set IP
    # This can happen if a device has multiple Ethernet interfaces like a Wired, Wifi, or VPN
    # This message lets user know which is bound
    if hostIP != IPaddr :
        #logging.warning("MAIN: *** Server_IP(%s) does not match discovered IP(%s) *** ", PigSettings['Server_IP'],IPaddr )
        logging.warning("MAIN: *** Server_IP(%s) does not match discovered IP(%s) *** ", hostIP, IPaddr )


#====================================================================================
# Load state of Past DVRs to provide continuity
# ToDo should provide a method to clean this out if a DVR has not been seen for a while

if 'DVR_States' in PigSettings :
    logging.info("MAIN: Loading DVR_States...")
    LS_SetDictOfDVRs( PigSettings['DVR_States'] )
else :
    logging.info("MAIN: No DVR_States to load")

#====================================================================================
# version from http://pigs.goihl.com/zips/version.txt to print update msg 1.36
# This check is to make sure pigs.json is up-to-date for the UI

CheckForPigsUpdate("MAIN:")
            
#====================================================================================
# Initialize the Alt server name to reach CM servers for CM/Rovi Icons
# CMInitAltServer()  now done in builder

sdapi = PigSettings["sdApiVersion"]

if args.api :   # check for cmd line over-ride
    sdapi = args.api 

setapi = SDSetAPIVersion(sdapi)  # set or display SD API

logging.info("MAIN: Setting SD API Version to '%s'", setapi)

#https://ywnd0zbafe.execute-api.us-east-1.amazonaws.com/20141201

#====================================================================================
# Start the Builder Thread
logging.info("MAIN: Starting BuilderThread")

#BuilderThread = threading.Thread(target=Builder, name='Builder', daemon=True, args=(PigSettings, PiGS_SERVICEID, BuildAbortEvent,) )

BuilderThread = threading.Thread(target=Builder, name='Builder', daemon=True )
BuilderThread.start()
logging.info("MAIN: BuilderThread.isalive(): %s", BuilderThread.is_alive())


#====================================================================================
# Load Programs and Cache from disk.  Do before DNS Start

from ls import LSLoadProgramsFromDisk   
from sv import SVLoadServicesFromDisk

if not args.noload :  # if its not noload, then it means load
    logging.info('MAIN: Loading Program Schedule from Disk...')
    SVLoadServicesFromDisk() # 1.40 Service list now saved and loaded
    LSLoadProgramsFromDisk()
else : 
    logging.info('MAIN: -n Skipping Program Schedule Load from Disk...')


#====================================================================================
# Start the DNS Server Thread now or delay until after first build?
# PigsDelayDNS causes DNS to start after Sched build completes
# This ensures that a DVR wont ask for schedule data before it is ready
# in 1.40 we start DNS immediately if a sched was loaded from disk

if PigSettings["PigsDoesDNS"] :
    
    # Need to force holdoff even if a schedule was loaded?
    # Typically used to sync ALL DVRs to a set build time
    # eg python3 pigs.py -delay 30 -holdoff
    if args.holdoff :
        logging.info("MAIN: DNS holdoff until after first build.")
        
    # If requesting delay, or there is no schedule, 
    elif PigSettings["PigsDelayDNS"] and not LSGetProgramSchedule():
        logging.info("MAIN: Delaying DNS Start until Schedule is Built.")

    # If a schedule was loaded from disk, we start DNS
    else :
        if LSGetProgramSchedule() :
            logging.info("MAIN: Schedule loaded from disk.  Queuing Start DNS...")
        # queue a 4s delay to let initialization complete
        BuilderQueue.put( {'Command':BuildDelay, 'Params':4} )
        BuilderQueue.put( {'Command':BuildStartDNS} )


#====================================================================================
# Build/Load the Service details so they are available to the Schedule builder.
# On very first run there is no user name so skip Initialization steps.
# It will get built when user does a Save Settings in the UI
# Also, if we never got an IP address, dont attempt to build the Service or Schedule lists

fbd = PigSettings['FirstBuildDelay']

if args.delay :   # check for cmd line First Build Delay (hrs)
    fbd = int(args.delay * 60 * 60)  # fbd in secs.  convert hrs to secs

# If its going to be a long time for the first build, and we have a loaded schedule,
# then initiate a build of everything except the schedule so the services list is ready?
# The Services list would be built if needed

#xxx todo if service list is loaded do I have to do this?

#if fbd > 300 and LSGetProgramSchedule() :
#    logging.info("MAIN: Delayed First Build. Queuing Initialization w/o Build...")
#    BuilderQueue.put( {'Command':BuildDelay, 'Params':2} )
#    BuilderQueue.put( {'Command':BuildInit } )
   
if IPaddr :
    
    if PigSettings['sdusername'] :

        ResetReadyState()
        
        # Build will do an Init to:
        # 0) Set the CM Alt Server url
        # 1) Verify SD account lineups match what user configured by deleting and adding lineups
        # 2) Build the Service list (Lineup list)
        # 3) Build the schedule, then enable DNS
        
        # Let webserver startup - Schedule the build a number of seconds from now
        if fbd :
            logging.info("MAIN: Initial Schedule Build in %s secs", fbd )
            fbt = SetDvrBuildTimer( fbd )
        else :
            logging.info("MAIN: Initial Schedule Build dsiabled.")
        
    else :
        logging.warning("MAIN: No SD username, can't Build Schedule! Configure at %s", 'http://'+IPaddr)
else :
    logging.warning("MAIN: PiGS could not determine IP address!  No Internet?")

    # set build timer to retry later.  The builder will attempt to find the IP address
    if fbd :
        SetDvrBuildTimer( fbd )  
    else :
        logging.info("MAIN: Initial Schedule Build dsiabled.")


#====================================================================================
# Construct and start the webserver

# References for HttpServer and SocketServer objects
#https://github.com/python/cpython/blob/3.10/Lib/socketserver.py
#https://github.com/python/cpython/blob/3.8/Lib/http/server.py
#https://github.com/python/cpython/blob/3.10/Lib/socketserver.py

if PigSettings["ThreadedHTTP"] :
    from http.server import ThreadingHTTPServer
else :
    from http.server import HTTPServer


class PiGS_HTTPServer( HTTPServer ):

    # testing of httpserver,socketserver for 1.24,1.30 lockups
    # override/extend TCPServer in socketserver for debug
    def XXget_request(self):
        print('getting...')
        rq = self.socket.accept()
        print('got ', rq)
        return rq
    
    # override/extend BaseServer in socketserver for debug
    def XXshutdown_request(self, request):
        print('closing ', request)
        self.close_request(request)
        
    # override to see if socketserver is polling (it does)
    def service_actions(self) :

        global gExpectedBuildTimeUTC
        global gLastPollTimeUTC
        
        dtn = datetime.utcnow()

        # If PiGS host PC wakes from sleep, display when and how long
        if PigSettings['WakeNotify'] :
            if gLastPollTimeUTC :
                pollDelta = dtn - gLastPollTimeUTC
                if pollDelta.total_seconds() > 120 :   #120
                    lts, tz = LocalTimeStr(dtn, showSeconds=True)
                    logging.warning("SVC_Actions: Woke from a %s sleep at %s ", str(pollDelta), lts)
            gLastPollTimeUTC = dtn

        # If the Host computer sleeps, the Build Timer Thread loses time and wont fire on time
        # Using webserver service_actions() polling to monitor the realtime clock 
        # and fire off the build if the expected build time has passed
        
        if PigSettings['SleepMonitor'] :
            #print( 'checking Build time', dtn, gExpectedBuildTimeUTC)
            if gExpectedBuildTimeUTC :
                if dtn > gExpectedBuildTimeUTC :
                    logging.warning("SVC_Actions Sleep Monitor: Expected Build Time passed!  Starting Build... ")
                    SetDvrBuildTimer(0.1) # Trigger build process 0.1 second from now
        
        #how to access dunder private class variables in a base class
        #print( self.server_address )
        #print( self.RequestHandlerClass )
        #print( self._BaseServer__shutdown_request )
        #print( self._BaseServer__is_shut_down.is_set() )
        #print( 'service_actions() %s %s', httpd._BaseServer__shutdown_request, httpd._BaseServer__is_shut_down.is_set() )

        HTTPServer.service_actions(self)  # call parent  


#====================================================================================
# create/bind instance of a web server object with my custom handler
Handler = PiGS_HTTPRequestHandler

RestartServerRequest = True  # initial value  

#httpd = None  # webserver handle
            
gWebServerRestartTime,tz = LocalTimeStr( PigsStartTime, True )

# Restart server loop
while RestartServerRequest :

    # Loop ends unless webserver requested to restart
    RestartServerRequest = False
    
    try :
        if PigSettings["ThreadedHTTP"] :
            logging.info('MAIN: ThreadingHTTPServer binding to IP=%s:%s',hostIP,port)
            
            # Change to threading server in 1.24 to handle Chrome favicon.ico requests
            httpd   = ThreadingHTTPServer((hostIP, port), Handler)
            # This class is identical to HTTPServer but uses threads to handle requests by using
            # the ThreadingMixIn. This is useful to handle web browsers pre-opening sockets, 
            # on which HTTPServer would wait indefinitely
        else :
            logging.info('MAIN: Non-threaded HTTPServer binding to IP=%s:%s',hostIP, port)
            #httpd   = HTTPServer((hostIP, port), Handler)
            httpd   = PiGS_HTTPServer((hostIP, port), Handler)
        
    except:
        logging.warning("MAIN: Could not bind Webserver to port %s. Another PiGS or Process runninng?", port )
        logging.warning("  On Linux did you start with root privs?  'sudo python3 pigs.py'?" )
        logging.warning("  On Linux try 'sudo lsof -i :80' to get Process ID #" )
        logging.warning("  On Linux Mint try sudo python3 pigs.py -s [pigs-IP-on-UI]  eg -s 192.168.1.200" )
        logging.warning("  On Windows, check firewall.  Allow PiGS full access on ports 53 & 80" )
        
        # Exit pigs, no sense in running pigs if we cant get the requested port for UI/DVR, usually 80
        sys.exit(5)

    logging.info("MAIN: Waiting for DVR+ requests on Port %d!\n",port)

    WebserverResetEvent.set()  # Signal initiator Reset has occurred

    # Start webserver... server_forever() does not end.
    # This is the 'main' thread... But allow for clean exit from keyboard with Ctrl-C
    # or from code (another thread) httpd.shutdown() can be used to break
    
    try:
        httpd.serve_forever(poll_interval=0.5)
    except KeyboardInterrupt:
        logging.info("MAIN: Ctrl-C Keyboard Interrupt. Webserver stopped!\n")
        pass

    httpd.server_close()

    httpd = None
    
    if RestartServerRequest :
        logging.info("\n******** MAIN: Restarting WebServer ********")
        gWebServerRestartTime, tz = LocalTimeStr( datetime.utcnow(), True )

        alertmsg = "PiGS WebServer not Responding!  Restarting...\r\n ." 
        SendEmailAlert(alertmsg, subjectMsg=': Webserver')
    
    # End of while RestartServerRequest : LOOP

# To exit, code will fall thru to here on a Ctrl-C

#====================================================================================
#Shut down Builder thread

logging.critical('\nMAIN: Shutting down threads and saving environment.')

if not SDResourceLock.acquire(False) :  # check if thread is busy
    logging.warning("MAIN: PiGS is busy building schedule.  Requesting Abort...")
    BuildAbortEvent.set()   #signal BuildLinearSchedule() to abort
    waitAttempts = 50 # 50 x 1000mS = 5 seconds
    while waitAttempts and not SDResourceLock.acquire(False) :
        time.sleep(0.1)
        waitAttempts -= 1
    BuildAbortEvent.clear() # ensure flag is clear after abort

# We have the lock(?), tell the Builder thread to terminate so we can exit nicely
logging.info("MAIN: Queuing BuildTerminate...")
BuilderQueue.put( {'Command':BuildTerminate} )
BuilderThread.join(2.0)  # give it a chance to end

# Shut down DNS Server Thread if alive
if PigsDnsAlive() :
    logging.info("MAIN: Stopping DNS.  Waiting for thread to end...")
    dnsThread = PigsDnsStop()
    dnsThread.join(2.0)  # wait up to 2 seconds for DNS to Stop (should stop in less than 1 sec)

# Save Settings on exit to preserve build times 
logging.info("MAIN: Saving DVR_States...")
PigSettings['DVR_States'] = LS_GetDictOfDVRs()

logging.critical('MAIN: Saving User Settings...')
SaveSettingsFile(PigSettings, cfgf)

#if PigSettings['SaveSchedule'] > 0:
#    logging.critical('MAIN: Saving Program Schedule to Disk...')
#    LSSaveProgramsToDisk( SaveLevel = PigSettings['SaveSchedule'] )

lst, lz = LocalTimeStr( datetime.utcnow(), WeekDay=False )
logging.critical("MAIN: PiGS done at %s.\n", lst)
