# se.py
#
# Searching Functions to support PiGS application
#
# This is a web browser search command
#
# Copyright October 2020, Timothy Goihl
#
# http://192.168.1.200/search?expand=&duration=&startdate=&zulu=&channel=&title=&copy=
#     title=           # string to find in Title (case insensistive)
#     copy=            # string to find in description Copy (case insensistive)
#     channel=         # only display listings for specific channel 3.1
#     expand=false     # true = show full description copy. Default false, about 100 chars
#     startdate=       # starting time format must be exact  2020-10-31T00:00:00Z
#     duration=        # Max minutes to seach forward.  Math works eg 2*24*60 = 2 days
#     zulu=false       # output results in Zulu time as received, false = local time
#     limit=N          # Limit to N results. Default to 100 results, 0=All results
#     pid=false        # Display the SD ProgramID
#     new=false        # =true will only find programs with the New flag set
#
# This is an AND search.  All conditions specified must be true to be a match
# Copy can be used to find episode names or number, actor names, etc.
# for Title & Copy search, its a sub-string match, not a word-match
#   so "nypd" finds nypd blue, but "Nypd  blue" does not (2 spaces)
#
# PiGS internal schedule Search Examples
# Display the first 60 minutes of the PiGS internal schedule
#    /search?expand=&duration=60&startdate=&zulu=true&channel=&title=&copy=
# Display full schedule for channel 3.1 
#    /search?expand=&duration=&startdate=&zulu=true&channel=3.1&title=&copy=
#    /search?channel=3.1
# Find all programs with the word 'fire' in the description
#    /search?expand=&duration=&startdate=&zulu=true&channel=&title=&copy=fire
# Actor search: find all programs with the word 'franz' in the description
#    /search?expand=&duration=&startdate=&zulu=true&channel=&title=&copy=franz
#    /search?copy=franz
# Show name search: Find nypd blue (just search nypd)
#    /search?expand=&duration=&startdate=&zulu=true&channel=&title=nypd&copy=
#    /search?title=nypd
# Limit name search to DST change over 10/31 to 11/1 for channel 15.1 
#    /search?expand=&duration=2*24*60&startdate=2020-10-31T00:00:00Z&zulu=true&channel=15.1&title=&copy=
#     start 10/31, for 2 days (in minutes) = 2*24*60
#    /search?startdate=2020-10-31T00:00:00Z&duration=2*24*60channel=15.1

SRCH_LIMIT = 100

from datetime import timedelta, datetime
from dateutil import parser
from ut import LocalTimeStr
import logging

def pageHeader(offset=0) :
    
    htmlhead  = "<html><head><title>Search Results</title>"
        #table.center {
        #  margin-left: auto; 
        #  margin-right: auto;
        #}
    # Style section of Header
    htmlhead += """
        <style>
        table, th, td {
          border: 1px solid black;
        }
        th { text-align: left; }
        </style>"""
    htmlhead += "</head><body><p>"  # end of header
    
    # Table Def and Table header row
    htmlhead += '<table>'
    htmlhead += '''<tr>
        <th style="width:50px">Chan</th>
        <th style="width:165px">Date-Time(offset={})</th>
        <th style="width:50px">Dura</th>
        <th style="width:275px">Title</th>
        <th >Description</th></tr>'''.format(offset)

#    htmlhead += '''<tr>
#        <th style="width:50px">Chan</th>
#        <th style="width:165px">Date-Time(offset={})</th>
#        <th style="width:50px">Dura</th>
#        <th style="width:50px">Res</th>
#        <th style="width:275px">Title</th>
#        <th >Description</th></tr>'''.format(offset)
    
    return htmlhead

def tableFooter() :
    return "</table></p>"

def pageFooter() :
    return "</body></html>"


#http://192.168.1.200/search?channel=3.1&title=ncis&copy=dennis

def SearchSchedule(mal, querystr) :
    
    # parse the params

    # Figure out search request window start and end times in Zulu
    rsts = mal[0]['AiringTime'] # default start time to first item in list
    
    # if user has specified a start date
    if 'startdate' in querystr : rsts = querystr['startdate'][0]    
    rst = parser.parse( rsts )

    # Start at current time minus a little bit (v1.40)
    from datetime import timezone
    bStartNow = querystr['startnow'][0].lower() =='true' if 'startnow' in querystr else False     # display new programs only

    if bStartNow :
        rst = datetime.now(timezone.utc) - timedelta(hours = 2)
    
    dur = 24*60*21   # default to 21 days in minutes = 30240 mintues
    if 'duration' in querystr : dur = int(eval(querystr["duration"][0]))    # allow math 21*24*60
    ret = rst + timedelta(minutes=dur)

    limit    = int(querystr['limit'][0]) if 'limit' in querystr else SRCH_LIMIT   # 0 means all results
    channel  = querystr['channel'][0] if 'channel' in querystr else None          # channel filter start time string
    titlestr = querystr['title'][0].lower() if 'title' in querystr else None      # Title string or None
    copystr  = querystr['copy'][0].lower() if 'copy' in querystr else None        # Description Copy
    Expand   = querystr['expand'][0].lower() =='true' if 'expand' in querystr else False  # Expand to show full description
    ZuluTime = querystr['zulu'][0].lower() =='true' if 'zulu' in querystr else False   # default local time, false for local time
    showpid  = querystr['pid'][0].lower() =='true' if 'pid' in querystr else False     # append PID in Description
    newprogs = querystr['new'][0].lower() =='true' if 'new' in querystr else False     # display new programs only
    
    logging.info("SEARCH: Ch=%s Title=%s Copy=%s Dur=%s Lim=%s New=%s Zulu=%s", \
                     channel,titlestr, copystr, dur,limit,newprogs,ZuluTime)

    # get timezone offset tz in minutes
    tt,tz = LocalTimeStr(rst, showSeconds=True, showDate=True, showTimeZone=False)
    if ZuluTime : tz = 0
    
    resultcnt = 0  # count how many matches we find
    searchStart = datetime.utcnow()
    
    # start building the webpage
    htmlpage = pageHeader(tz)

    for a in mal :  # scan all listings
        
        #get program airing start time STRING like "2020-03-08T08:00:00Z"
        aa = pss = a['AiringTime']
        pst = parser.parse( pss )      # convert to a TIME - prog start time pst
        
        # if the Program Start time pst is before the requested start time rst, skip it
        if (pst < rst) : continue      # to next listing
        
        # if the Progam Start time pst is after the ret, we are done
        if (pst > ret) : break         # out of loop

        # Now test listings - this is an AND test, all must be true
        
        # If an item is specified but does not match the listing, continue on to the next
        if channel and channel != a['Channel'] :
            continue
        if titlestr and titlestr not in a['Title'].lower() :
            continue
        if copystr and copystr not in a['Copy'].lower() :
            continue        
        if newprogs and a["AiringType"] != "New" :
            continue
    
        # all conditions matched, output the listing

        resultcnt += 1   # found a match - count it and print it
        
        if not ZuluTime :  # create a local time string
            aa,tz = LocalTimeStr(pst, showSeconds=True, showDate=True, showTimeZone=False, WeekDay=True) 
        ad = a['Duration']
        an = a['Channel']
        at = a['Title']
        ar = a['HDLevel']
        ac = a['Copy'] if Expand else a['Copy'][:100]
        if showpid :
            ac = a['ProgramId'] + '.  ' + ac 
        
        # Format a table row
        rowf ='<tr> <td style="text-align:center">{}</td> <td>{}</td> <td style="text-align:center">{}</td> <td>{}</td> <td>{}</td> </tr>'
        rowstr = rowf.format(an,aa,ad,at,ac)
        #rowf ='<tr> <td style="text-align:center">{}</td> <td>{}</td> <td style="text-align:center">{}</td> <td style="text-align:center">{}</td> <td>{}</td> <td>{}</td> </tr>'
        #rowstr = rowf.format(an,aa,ad,ar,at,ac)
        htmlpage += rowstr  # append row to html page
        
        # stop if we reach users result limit. if limit is 0, it wont stop 
        if resultcnt == limit : break

    htmlpage += tableFooter()

    tdiff = datetime.utcnow()-searchStart
    ts=str.format('{:02}:{:02}s', tdiff.seconds//60, tdiff.seconds%60)
    
    endstr    = "<BR>End of Search. {} Results.  {} <BR>  ".format(resultcnt,ts)
    htmlpage += endstr 
    
    if resultcnt == limit :
        endstr    = "<BR>Search Limit reached! There may be more results.  Use &limit=0 or &limit=N<BR>"
        htmlpage += endstr
        
#    if 'help' in querystr.lower() :   # print some help
#        helpstr    = "<BR>Usage: Dispaly Help"
#        htmlpage += helpstr
    
    htmlpage += pageFooter()
    
    return htmlpage




def lineupHeader() :
    
    htmlhead  = "<html><head><title>PiGS Lineup</title>"
        #table.center {
        #  margin-left: auto; 
        #  margin-right: auto;
        #}
    # Style section of Header
    htmlhead += """
        <style>
        table, th, td { border: 1px solid black; }
        th { text-align: left; }
        </style>"""
    htmlhead += "</head><body><p>"  # end of header
    
    # Table Def and Table header row
    htmlhead += '<table>'
    htmlhead += '''<tr>
        <th style="width:50px">VChan</th>
        <th style="width:50px">RF</th>
        <th style="width:100px">Call Sign</th>
        <th style="width:100px">Affiliate</th>
        <th style="width:75px">Stn ID</th>
        <th style="width:125px">Lineup</th>
        <th style="width:400px">Icon Link</th></tr>'''
    
    return htmlhead


def DisplayLineup(LineupList, querystr) :
        
    resultcnt = 0  # count how many matches we find
    
    # start building the webpage
    htmlpage = lineupHeader()

    for svc in LineupList :  # scan the Lineup

        resultcnt += 1
        
        #[VChan][RF][CallSn][Affil][StaID][IconLink]
        ch = svc["Channel"]
        rf = svc['ChannelLocation']
        cs = svc["DisplayName"] 
        af = svc["CallLetters"]
        st = svc['SourceId']
        lu = svc["SDLineupName"]
        url= svc["ChannelImages"][0]["ImageUrl"]
        
        # Format a table row
        rowf ='<tr> <td style="text-align:center">{}</td> \
                    <td style="text-align:center">{}</td> \
                    <td style="text-align:center">{}</td> \
                    <td style="text-align:center">{}</td> \
                    <td style="text-align:center">{}</td> \
                    <td style="text-align:center">{}</td> <td>{}</td></tr>'
        rowstr = rowf.format(ch,rf,cs,af,st,lu,url)
        htmlpage += rowstr  # append row to html page
        
    htmlpage += tableFooter()

    endstr    = "<BR>  {} channels in Lineup.  <BR>  ".format(resultcnt)
    htmlpage += endstr 
    
    htmlpage += pageFooter()
    
    return htmlpage

